#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import Sequence, Union, Optional, Callable, Tuple, Any, List, Dict, Type

import functools, logging, subprocess, sys, os
from pathlib import Path

from PySide2.QtCore import QObject, QProcess, Signal
from PySide2.QtWidgets import QMessageBox, QApplication

from src import mg_config
from src.mg_utils import hasGitAuthFailureMsg
import src.mg_const as mg_const
from src.mg_auth_failure_mgr import MgAuthFailureMgr

logger = logging.getLogger('mg_tools')
dbg = logger.debug
warn = logger.warning
log_git_cmd = logging.getLogger(mg_const.LOGGER_GIT_CMD).info

GIT_EXIT_CODE_SEGFAULT_OF_GIT_WITH_STACKTRACE = -1
GIT_EXIT_CODE_SEGFAULT_OF_GIT_NO_STACKTRACE = -2
GIT_EXIT_CODE_COULD_NOT_START_PROCESS = -3
GIT_EXIT_CODE_STOPPED_BECAUSE_AUTH_FAILURE = -4


def isRunningInsideSnap() -> bool:
    """Return TTrue if running inside a snap container.

    Used to identify the proper program locations which are different when running inside a snap."""
    return 'SNAP_EUID' in os.environ


def snapRoot() -> Path:
    '''Return the equivalent of root under snap, used to look for /usr/bin for example'''
    ret = Path(os.environ.get('SNAP', '/'))
    return ret



class ExecTool:
    # List of platform (as returned by sys.platform()) where we can run this tool
    SUPPORTED_PLATFORMS: List[str]

    WIN32_PATH_CANDIDATES: List[Path] = []
    LINUX_PATH_CANDIDATES: List[Path] = []

    EXEC_NAME_LINUX: str
    EXEC_NAME_WIN32: str

    # if not None, this is a command which we can run to detect the program. Typically, this is ['--version']
    INNOCUOUS_COMMAND: Optional[List[str]] = None

    # name of the configuration entry to store auto-detection behavior
    CONFIG_ENTRY_AUTODETECT: str

    # name of the configuration entry to store the manual path to the program
    CONFIG_ENTRY_MANUAL_PATH: str

    # name of the configuration entry to store if the program is activated or not
    CONFIG_ENTRY_ACTIVATED: str

    DOUBLE_CLICK_ACTIONS: List[str] = []

    SESSION_CACHE: Dict[Type['ExecTool'], str] = {
    }


    @classmethod
    def platform_supported(cls) -> bool:
        '''Return whether the current platform is supported'''
        return sys.platform in cls.SUPPORTED_PLATFORMS


    @classmethod
    def autodetect_executable(cls) -> str:
        '''Autodetect the executable location according to a few heuristics and return the result.
        Returns an empty string when nothing found.
        '''
        if not cls.platform_supported():
            return ''

        # check if we have it in cache
        if cls in cls.SESSION_CACHE:
            return cls.SESSION_CACHE[cls]

        if cls.INNOCUOUS_COMMAND:
            # first, try on the command-line
            try:
                exit_code, output = RunProcess().exec_blocking([cls.get_exec_name()] + cls.INNOCUOUS_COMMAND, allow_errors=True)
                if exit_code == 0:
                    # program is on the path, use it
                    cls.SESSION_CACHE[cls] = cls.get_exec_name()
                    return cls.get_exec_name()
            except FileNotFoundError:
                # program is not on the path
                pass


        # Look in some standards locations
        if sys.platform == 'win32':
            path_candidates = cls.WIN32_PATH_CANDIDATES
        elif sys.platform == 'linux':
            if isRunningInsideSnap():
                # to find the path of a file inside the snap, remove the first '/'
                # and make the path relative to snap root
                path_candidates = [ snapRoot() / str(p)[1:] for p in cls.LINUX_PATH_CANDIDATES ]
            else:
                path_candidates = cls.LINUX_PATH_CANDIDATES
        else:
            raise ValueError('Unsupported platform')

        cmd = cls.find_prog_exec(path_candidates)

        # put it in cash for next time
        if cmd:
            cls.SESSION_CACHE[cls] = cmd

        # return the result even if we could not find anything
        return cmd


    @classmethod
    def get_executable(cls) -> str:
        '''Check if the configuration `autodetect` is True or None and in this case,
        auto-detects the program location. Else, simply return the value defined in `config_entry_manual_path`
        '''
        if not cls.platform_supported():
            return ''

        config = mg_config.get_config_instance()
        if config[cls.CONFIG_ENTRY_AUTODETECT] in (None, True):
            result = cls.autodetect_executable()
        else:
            result = config[cls.CONFIG_ENTRY_MANUAL_PATH]
        return result


    @classmethod
    def get_exec_name(cls) -> str:
        if sys.platform == 'win32':
            return cls.EXEC_NAME_WIN32
        elif sys.platform == 'linux':
            return cls.EXEC_NAME_LINUX
        else:
            raise ValueError('Platform not supported: ', sys.platform)


    @classmethod
    def find_prog_exec(cls, path_candidates: Sequence[Union[str, Path]]) -> str:
        '''Scans locations path_candidates to see if the expected program exists. Return the program at this location if found or an empty string.

        The path_candidates are a list of directory location where one can find exec_name

        Example:
            path_candidates: List[Path] = [
                Path(os.environ["ProgramFiles(x86)"])/"Git"/"bin",
                Path(os.environ["ProgramFiles"])/"Git"/"bin",
            ]

            find_prog_exec( path_candidates )
        '''
        dbg(f'find_prog_exec({cls}, {path_candidates})')
        exec_name = cls.get_exec_name()
        dbg(f'find_prog_exec() - exec_name={exec_name}')
        for possible_path in path_candidates:
            candidate_path = Path(possible_path) / exec_name
            dbg('Looking at: {}'.format(str(candidate_path)))
            if not candidate_path.exists():
                continue
            # ok, we have it !
            dbg('Found program at: {}'.format(str(candidate_path)))
            return str(candidate_path)

        # could not find the wanted program
        return ''


    @classmethod
    def shouldShow(cls) -> bool:
        '''Return whether to show the program in menu:
        - for first run, we show if the program is autodetected on the computer
        - for non first run, we show according to config
        '''
        if mg_config.get_config_instance().get(cls.CONFIG_ENTRY_ACTIVATED) is None:
            # configuration entry does not exist, this is our first run
            if cls.get_executable():
                # program is not configured and but is autodetected
                showProgram = True
            else:
                # program is not configured and not autodetected
                showProgram = False
            mg_config.get_config_instance()[cls.CONFIG_ENTRY_ACTIVATED] = showProgram
        else:
            showProgram = mg_config.get_config_instance().get(cls.CONFIG_ENTRY_ACTIVATED)
        return showProgram


    @classmethod
    def exec_non_blocking(cls, cmd_args: List[str], workdir: str = '', allow_errors: bool = False, callback: Optional[Callable[[], Any]] = None) -> None:
        '''Run the program with the arguments listed in cmd_args, in the working directory.
        Raises an exception if the command did not return with 0 status and allow_errors is False.
        '''
        prog_to_run = cls.get_executable()
        if prog_to_run == '':
            raise FileNotFoundError(f'Could not locate {cls.get_exec_name()}')

        cb_done = None
        if callback is not None:
            def cb_done(_exit_code: int, _output: str) -> None:
                callback()

        rp = RunProcess()
        cmdline = [prog_to_run] + cmd_args
        rp.exec_async(cmdline=cmdline, cb_done=cb_done, working_dir=workdir, allow_errors=allow_errors)


    @classmethod
    def exec_blocking(cls, cmd_args: Sequence[str], workdir: str = '', allow_errors: bool = False) -> str:
        '''Run the program with the arguments listed in cmd_args, in the working directory.
        Raises an exception if the command did not return with 0 status.
        '''
        prog_to_run = cls.get_executable()
        if prog_to_run == '':
            raise FileNotFoundError(f'Could not locate {cls.get_exec_name()}')

        cmdline = [prog_to_run] + list(cmd_args)
        exitCode, output = RunProcess().exec_blocking(cmdline, allow_errors=allow_errors, working_dir=workdir)
        if exitCode != 0 and not allow_errors:
            raise subprocess.CalledProcessError(cmd=cmdline, returncode=exitCode, output=output)
        return output


    @classmethod
    def doubleClickActions(cls) -> List[str]:
        '''Return the double click actions provided by this tool on the current platform'''
        if not cls.platform_supported() or not cls.shouldShow():
            return []

        return cls.DOUBLE_CLICK_ACTIONS



#######################################################
#       Git Stuff
#######################################################

class ExecGit(ExecTool):
    SUPPORTED_PLATFORMS = ['win32', 'linux']

    WIN32_PATH_CANDIDATES: List[Path] = [
        Path(os.environ.get("ProgramFiles", '')) / "Git" / "bin",
        Path(os.environ.get("ProgramFiles", '')) / "Git" / "cmd",
        Path(os.environ.get("ProgramFiles(x86)", '')) / "Git" / "bin",
        Path(os.environ.get("ProgramFiles(x86)", '')) / "Git" / "cmd",
        Path(os.environ.get("PROGRAMW6432", '')) / "Git" / "bin",
        Path(os.environ.get("PROGRAMW6432", '')) / "Git" / "cmd",
    ]

    LINUX_PATH_CANDIDATES = [
        Path('/usr/bin'),
        Path('/usr/local/bin'),
    ]

    EXEC_NAME_LINUX = 'git'
    EXEC_NAME_WIN32 = 'git.exe'

    INNOCUOUS_COMMAND = ['--version']

    CONFIG_ENTRY_AUTODETECT = mg_config.CONFIG_GIT_AUTODETECT
    CONFIG_ENTRY_MANUAL_PATH = mg_config.CONFIG_GIT_MANUAL_PATH

    @classmethod
    def checkFound(cls) -> bool:
        '''If the explorer program is not found, display an error dialog and return False.

        Else, simply return True
        '''
        if cls.get_executable() == '':
            QMessageBox.warning(None, 'Could not execute git', 'Error: could not execute git\n'
                                '\nMultiGit needs git to work.\nPlease configure the location in the preference dialog.\n')
            return False

        return True


#######################################################
#       TortoiseGit Stuff
#######################################################

class ExecTortoiseGit(ExecTool):
    SUPPORTED_PLATFORMS = ['win32']

    WIN32_PATH_CANDIDATES = [
        Path(os.environ.get("ProgramFiles(x86)", '')) / "TortoiseGit"/"bin",
        Path(os.environ.get("PROGRAMW6432", '')) / "TortoiseGit"/"bin",
        Path(os.environ.get("ProgramFiles", '')) / "TortoiseGit"/"bin",
    ]

    EXEC_NAME_WIN32 = "TortoiseGitProc.exe"

    CONFIG_ENTRY_AUTODETECT = mg_config.CONFIG_TORTOISEGIT_AUTODETECT
    CONFIG_ENTRY_MANUAL_PATH = mg_config.CONFIG_TORTOISEGIT_MANUAL_PATH
    CONFIG_ENTRY_ACTIVATED = mg_config.CONFIG_TORTOISEGIT_ACTIVATED

    DOUBLE_CLICK_ACTIONS = [
        mg_const.DBC_TORTOISEGITSHOWLOG,
        mg_const.DBC_TORTOISEGITCOMMIT,
        mg_const.DBC_TORTOISEGITSWITCH,
        mg_const.DBC_TORTOISEGITBRANCH,
        mg_const.DBC_TORTOISEGITPUSH,
        mg_const.DBC_TORTOISEGITPULL,
        mg_const.DBC_TORTOISEGITFETCH,
        mg_const.DBC_TORTOISEGITDIFF,
    ]



#######################################################
#       SourceTree stuff
#######################################################

class ExecSourceTree(ExecTool):
    SUPPORTED_PLATFORMS = ['win32']

    WIN32_PATH_CANDIDATES = [
        Path(os.environ.get("ProgramFiles(x86)", '')) / "Atlassian" / "SourceTree",
        Path(os.environ.get("PROGRAMW6432", '')) / "Atlassian" / "SourceTree",
        Path(os.environ.get("ProgramFiles", '')) / "Atlassian" / "SourceTree",
    ]

    EXEC_NAME_WIN32 = "SourceTree.exe"

    CONFIG_ENTRY_AUTODETECT = mg_config.CONFIG_SOURCETREE_AUTODETECT
    CONFIG_ENTRY_MANUAL_PATH = mg_config.CONFIG_SOURCETREE_MANUAL_PATH
    CONFIG_ENTRY_ACTIVATED = mg_config.CONFIG_SOURCETREE_ACTIVATED

    DOUBLE_CLICK_ACTIONS = [
        mg_const.DBC_RUNSOURCETREE,
    ]



#######################################################
#       SublimeMerge stuff
#######################################################

class ExecSublimeMerge(ExecTool):
    SUPPORTED_PLATFORMS = ['win32', 'linux']

    WIN32_PATH_CANDIDATES = [
        Path(os.environ.get("ProgramFiles", '')) / "Sublime Merge",
        Path(os.environ.get("PROGRAMW6432", '')) / "Sublime Merge",
    ]
    LINUX_PATH_CANDIDATES = [
        Path('/opt/sublime_merge'),
    ]

    EXEC_NAME_WIN32 = "sublime_merge.exe"
    EXEC_NAME_LINUX = "sublime_merge"

    CONFIG_ENTRY_AUTODETECT = mg_config.CONFIG_SUBLIMEMERGE_AUTODETECT
    CONFIG_ENTRY_MANUAL_PATH = mg_config.CONFIG_SUBLIMEMERGE_MANUAL_PATH
    CONFIG_ENTRY_ACTIVATED = mg_config.CONFIG_SUBLIMEMERGE_ACTIVATED

    DOUBLE_CLICK_ACTIONS = [
        mg_const.DBC_RUNSUBLIMEMERGE,
    ]

#######################################################
#       GitBash stuff
#######################################################

class ExecGitBash(ExecTool):
    SUPPORTED_PLATFORMS = ['win32']

    WIN32_PATH_CANDIDATES = [
        Path(os.environ.get("ProgramFiles", '')) / "Git",
        Path(os.environ.get("ProgramFiles(x86)", '')) / "Git",
        Path(os.environ.get("PROGRAMW6432", '')) / "Git",
        ]

    INNOCUOUS_COMMAND = ['--version']

    EXEC_NAME_WIN32 = "git-bash.exe"

    CONFIG_ENTRY_AUTODETECT = mg_config.CONFIG_GITBASH_AUTODETECT
    CONFIG_ENTRY_MANUAL_PATH = mg_config.CONFIG_GITBASH_MANUAL_PATH
    CONFIG_ENTRY_ACTIVATED = mg_config.CONFIG_GITBASH_ACTIVATED

    DOUBLE_CLICK_ACTIONS = [
        mg_const.DBC_RUNGITBASH,
    ]

#######################################################
#       GitGui stuff
#######################################################

class ExecGitGui(ExecTool):
    SUPPORTED_PLATFORMS = ['win32', 'linux']

    WIN32_PATH_CANDIDATES = [
        Path(os.environ.get("ProgramFiles", '')) / "Git" / "cmd",
        Path(os.environ.get("ProgramFiles(x86)", '')) / "Git" / "cmd",
        Path(os.environ.get("PROGRAMW6432", '')) / "Git" / "cmd",
        ]

    LINUX_PATH_CANDIDATES = [
        Path('/usr/lib/git-core'),  # Ubuntu
    ]

    EXEC_NAME_WIN32 = "git-gui.exe"
    EXEC_NAME_LINUX = "git-gui"

    CONFIG_ENTRY_AUTODETECT = mg_config.CONFIG_GITGUI_AUTODETECT
    CONFIG_ENTRY_MANUAL_PATH = mg_config.CONFIG_GITGUI_MANUAL_PATH
    CONFIG_ENTRY_ACTIVATED = mg_config.CONFIG_GITGUI_ACTIVATED

    DOUBLE_CLICK_ACTIONS = [
        mg_const.DBC_RUNGITGUI,
    ]

#######################################################
#       Gitk stuff
#######################################################

class ExecGitK(ExecTool):
    SUPPORTED_PLATFORMS = ['win32', 'linux']

    WIN32_PATH_CANDIDATES = [
        Path(os.environ.get("ProgramFiles", '')) / "Git" / "cmd",
        Path(os.environ.get("ProgramFiles(x86)", '')) / "Git" / "cmd",
        Path(os.environ.get("PROGRAMW6432", '')) / "Git" / "cmd",
    ]

    LINUX_PATH_CANDIDATES = [
        Path('/usr/bin'),  # Ubuntu
    ]

    EXEC_NAME_WIN32 = "gitk.exe"
    EXEC_NAME_LINUX = "gitk"

    CONFIG_ENTRY_AUTODETECT = mg_config.CONFIG_GITK_AUTODETECT
    CONFIG_ENTRY_MANUAL_PATH = mg_config.CONFIG_GITK_MANUAL_PATH
    CONFIG_ENTRY_ACTIVATED = mg_config.CONFIG_GITK_ACTIVATED

    DOUBLE_CLICK_ACTIONS = [
        mg_const.DBC_RUNGITK,
    ]

#######################################################
#       Open directory
#######################################################

class ExecExplorer(ExecTool):
    SUPPORTED_PLATFORMS = ['win32', 'linux']

    # no path candidates, it should be on the execution path
    WIN32_PATH_CANDIDATES = [
        Path(os.environ.get("SystemRoot", '')),
    ]
    LINUX_PATH_CANDIDATES = [
        Path("/usr/bin"),
    ]

    EXEC_NAME_WIN32 = 'explorer.exe'
    EXEC_NAME_LINUX = 'xdg-open'

    # name of the configuration entry to store auto-detection behavior
    CONFIG_ENTRY_AUTODETECT = mg_config.CONFIG_EXPLORER_AUTODETECT

    # name of the configuration entry to store the manual path to the program
    CONFIG_ENTRY_MANUAL_PATH = mg_config.CONFIG_EXPLORER_MANUAL_PATH

    @classmethod
    def checkFound(cls) -> bool:
        '''If the explorer program is not found, display an error dialog and return False.

        Else, simply return True
        '''
        if cls.get_executable() == '':
            QMessageBox.warning(None, 'Could not find the explorer program',
                                'Could not find a program to open directories. Please choose one in the File/Settings dialog.')
            return False

        return True

    @classmethod
    def exec_non_blocking(cls, cmd_args: List[str], workdir: str = '', allow_errors: Optional[bool] = None, callback: Optional[Callable[[], Any]] = None) -> None:
        '''Same as default exec_non_blocking but defaults to allow_errors for Windows'''
        if allow_errors == None:
            if sys.platform == 'win32':
                # on Windows, launching the explorer returns -1
                override_allow_errors = True
            else:
                override_allow_errors = False
        return super().exec_non_blocking(cmd_args, workdir, override_allow_errors, callback)


#######################################################
#       GitProcess class
#######################################################

FORCE_ASYNC_TO_BLOCKING_CALLS = False

class RunProcess(QObject):
    '''Utility class to run a git or another process synchronously or asynchronously.

    One RunProcess() only shall be created for a repository and it will ensure that only one git command
    is running at the same time on the repository.

    This class also uses the GitProcessPool to limit the number of total git processes running in parallel.

    Typical usage:
    ==============

    # aysnchronous git execution
    rp = RunProcess('some_repo')
    rp.exec_async(callback_git_operation_finished, ['-C', 'some_repo', 'fetch'])

    rp.exec_async(callback_git_operation_finished, ['-C', 'some_repo', 'fetch'], force_blocking=True)
    rp.exec_async(callback_git_operation_finished, ['-C', 'some_repo', 'fetch'], allow_errors=True)

    # synchronous git execution
    rp = RunProcess('some_repo')
    git_exit_code, cmd_out = rp.exec_blocking(None, ['-C', 'some_repo', 'fetch'])

    # asynchronous execution of another program
    rp = RunProcess(None)
    rp.process_working_dir = gitdir
    rp.cmd = prog_gitbash
    rp.exec_async(args=[], cb_done=cb_done)

    The interest of this class above QProcess:
    - limit one git execution over a given repository
    - detects some git execution errors corner cases
    - shows a dialog box if git executable can not be found
    - shows a dialog box if the executable returned an error (configurable with allow_errors)
    - asynchronous execution is slightly easier than when dealing directly with QProcess
    - the debug log captures partial stdout/stderr which eases debugging

    You can use the signal sigProcessOutput to track the progressive output of the process.

    '''

    process: Optional[QProcess]
    cb_done: Optional[ Callable[[int, str], Any]]
    process_working_dir: Optional[str]
    sigProcessOutput = Signal(str)

    def __init__(self) -> None:
        super().__init__(QApplication.instance())

        self.cmdline: Sequence[str] = []
        self.process_working_dir = None
        self.process = None
        self.cb_done = None
        self.partial_stdout = ''
        self.last_exit_code = 0
        self.allow_errors = False
        self.emit_output = False


    def nice_cmdline(self) -> str:
        return ' '.join(self.cmdline)


    def create_process(self, cmdline: Sequence[str], working_dir: str = '',
                       allow_errors: bool = False, emit_output: bool = False) -> None:
        '''Creates a QProcess for running a git command

        cmdline: argument to git to execute.

        May block while a git process is currently under execution
        '''
        self.cmdline = cmdline
        self.emit_output = emit_output
        self.process = QProcess(self)
        self.process.setProgram(cmdline[0])
        self.process.setArguments(cmdline[1:])
        if len(working_dir):
            self.process.setWorkingDirectory(working_dir)
        self.allow_errors = allow_errors

        self.last_exit_code = 0
        self.process.setProcessChannelMode(QProcess.ProcessChannelMode.MergedChannels)

        if self.emit_output:
            # noinspection PyUnresolvedReferences
            self.process.readyReadStandardOutput.connect(self.slotReadyReadStdout)


    def slotReadyReadStdout(self) -> None:
        '''Slot called when data is available from the process stdout channel'''
        assert self.process
        btext = bytes(self.process.readAllStandardOutput())
        self.partial_stdout += btext.decode('utf8', errors='replace')
        logger.debug('%s - partial git output:' % self.nice_cmdline())
        logger.debug(b'"%r"' % btext)
        if self.emit_output:
            self.sigProcessOutput.emit(self.partial_stdout)

        if hasGitAuthFailureMsg(self.partial_stdout):
            MgAuthFailureMgr.gitAuthFailed(self.nice_cmdline())


    def exec_blocking(self, cmdline: Sequence[str], allow_errors: bool = False, working_dir: str = '') -> Tuple[int, str]:
        '''Execute a git command in blocking mode.

        args: the arguments to git.

        Return the git exit code and the standard output of the command.

        If allow_errors is False (the default) a message box reports the git error. If it is True, nothing
        is done. In both cases, the git exit code is returned.
        '''
        self.create_process(cmdline, working_dir=working_dir, allow_errors=allow_errors)

        if MgAuthFailureMgr.shouldStopBecauseAuthFailureInProgress(cmdline):
            self.last_exit_code = GIT_EXIT_CODE_STOPPED_BECAUSE_AUTH_FAILURE
            self.partial_stdout = 'Stopped before execution because too many authentication failures.'
            dbg('Execution canceled because of too many authentication failures: {}'.format(self.nice_cmdline()))
            cmd_out = self.process_finished(self.last_exit_code, QProcess.ExitStatus.NormalExit)
            return self.last_exit_code, cmd_out

        assert self.process
        dbg('Executing blocking: {}'.format(self.nice_cmdline()))
        self.process.start()

        if (self.process.state() not in (QProcess.Starting, QProcess.Running)
                or not self.process.waitForStarted(-1)):
            # could not start the process...
            cmd_out = self.process_finished(GIT_EXIT_CODE_COULD_NOT_START_PROCESS, QProcess.ExitStatus.CrashExit)
        else:
            finished = self.process.waitForFinished(-1)
            if not finished:
                logger.warning('Process returned before being finished...')
                raise AssertionError('process not yet finished...')

            cmd_out = self.process_finished(self.process.exitCode(), self.process.exitStatus())

        return self.last_exit_code, cmd_out


    def exec_async(self, cmdline: Sequence[str],
                   cb_done: Optional[Callable[[int, str], Any]],
                   force_blocking: bool = False,
                   allow_errors: bool = False,
                   working_dir: str = '',
                   emit_output: bool = False) -> None:
        '''Execute a git command asynchronously and calls cb_git_done() when the command completes.

        cmdline: the full command-line including the program. When running over a repo, the first argument
                 of the command is replaced by the actual path to git, extracted from the config.

        Calls: cb_git_done(git_exit_code, git_output)

        if force_blocking is True, the call is made blocking and the callback is called
        when the git process completes.

        If allow_errors is False (the default) a message box reports the git error. If it is True, nothing
        is done. In both cases, the git exit code is passed to the callback.

        If emit_output is True, the signal sigProcessOutput is emitted progressively
        as the process emits output.
        '''
        self.cb_done = cb_done
        if FORCE_ASYNC_TO_BLOCKING_CALLS or force_blocking:
            self.exec_blocking(cmdline, allow_errors, working_dir)
            return

        self.create_process(cmdline, working_dir=working_dir, allow_errors=allow_errors,
                            emit_output=emit_output)

        assert self.process

        if MgAuthFailureMgr.shouldStopBecauseAuthFailureInProgress(cmdline):
            self.last_exit_code = GIT_EXIT_CODE_STOPPED_BECAUSE_AUTH_FAILURE
            self.partial_stdout = 'Stopped before execution because too many authentication failures.'
            dbg('Execution canceled because of too many authentication failures: {}'.format(self.nice_cmdline()))
            QApplication.processEvents()
            self.process_finished(self.last_exit_code, QProcess.ExitStatus.NormalExit)
            return

        # noinspection PyUnresolvedReferences
        self.process.finished.connect(self.process_finished)

        dbg('Executing async: {}'.format(self.nice_cmdline()))
        self.process.start()


    def process_finished(self, exit_code: int, exit_status: QProcess.ExitStatus) -> str:
        '''Called by both blocking and async process execution, when execution is over.

        Calls the cb_done if any.
        '''
        dbg('Process finished for "%s" with exit status %d, exit code %d' % (self.nice_cmdline(), int(exit_status), exit_code))
        assert self.process
        cmd_out: str
        btext = bytes(self.process.readAllStandardOutput())
        cmd_out = self.partial_stdout + btext.decode('utf8', errors='replace').replace('\r', '\n')
        self.partial_stdout = ''
        cmd_out_dbg = cmd_out if len(cmd_out) < mg_const.MAX_GIT_DBG_OUT_CHAR else \
            cmd_out[:mg_const.MAX_GIT_DBG_OUT_CHAR] + '\n<output truncated to {} characters>'.format(mg_const.MAX_GIT_DBG_OUT_CHAR)
        logger.debug('stdout: {}'.format(cmd_out_dbg))

        log_git_cmd(self.nice_cmdline())
        log_git_cmd('\t' + '\n\t'.join(cmd_out_dbg.split('\n')))

        # keep a local copy before final cleanup
        process = self.process
        cb_done = self.cb_done
        self.process = None
        self.cb_done = None

        if hasGitAuthFailureMsg(cmd_out_dbg):
            MgAuthFailureMgr.gitAuthFailed(self.nice_cmdline())

        self.last_exit_code = exit_code
        if  '** *fatal error - add_item' in cmd_out or 'Stack trace:' in cmd_out and self.last_exit_code == 0:
            # this happens sometimes with git process started very closely one to another with a fetch
            # that's why we use a delay between starting multiple git processes
            # usually, the error is not reported in git exit code, probably because it is not git-fetch
            # who is failing but a sub-program
            # so force exit code
            logger.error('Git segfault detected, forcing exit code to -1 for: %s' % self.nice_cmdline())
            self.last_exit_code = GIT_EXIT_CODE_SEGFAULT_OF_GIT_WITH_STACKTRACE

        elif process.exitStatus() != QProcess.NormalExit and self.last_exit_code == 0:
            logger.error('Git segfault detected, forcing exit code to -2 for: %s' % self.nice_cmdline())
            self.last_exit_code = GIT_EXIT_CODE_SEGFAULT_OF_GIT_NO_STACKTRACE

        if self.last_exit_code != 0 and not self.allow_errors:
            # noinspection PyTypeChecker
            if QApplication.instance():
                QMessageBox.warning(None, 'Error when running process',
                                    'Bad exit code %d, see command below:\n\n%s' % ( self.last_exit_code, self.nice_cmdline()))
            else:
                raise ValueError('Bad exit code %d for command: %s' % (self.last_exit_code, self.nice_cmdline()))

        # to avoid recursion, don't leave any pending callbacks
        if cb_done:
            logger.debug('Calling process done callback')
            cb_done(self.last_exit_code, cmd_out)

        return cmd_out


    def abortProcessInProgress(self) -> None:
        '''Abort the running git process.

        Does nothing if there is no process running
        '''
        if self.process:
            dbg('abortProcessInProgress() for %s, killing process ' % self.nice_cmdline())
            self.process.kill()
        else:
            dbg('abortProcessInProgress() for %s, no process to kill' % self.nice_cmdline())






#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List, Any, Optional, Dict, Callable, cast
import logging, subprocess

from PySide2.QtWidgets import QTreeWidget, QMenu, QAction, QApplication, QMessageBox, QAbstractItemView, QTreeWidgetItem
from PySide2.QtGui import QIcon
from PySide2.QtCore import Qt, QPoint

import src.mg_const as mg_const
from src import mg_config as mgc
from src.mg_tools import ExecTortoiseGit, ExecGitBash, ExecSourceTree, ExecSublimeMerge, ExecGit, ExecGitK, ExecExplorer, ExecGitGui
from src.mg_repo_info import MgRepoInfo
from src.mg_repo_tree_item import MgRepoTreeItem
from src.mg_exec_window import MgExecWindow
from src.mg_dialog_view_properties import runDialogGitProperties
from src.mg_dialog_settings import runDialogEditSettings
from src.mg_actions import MgActions
from src.mg_plugin_mgr import pluginMgrInstance

logger = logging.getLogger('mg_repo_tree')
dbg = logger.debug
warn = logger.warning

'''
What to keep in mg_window:
--------------------------
- Everything related to window menu management
- everything related to main window: last commit, last file, ...

What to put in mg_repo_tree:
----------------------------
tgui- operations specific to the repo being listed in the view
    + the rmb menu
    + operations related to selected items
'''


class MgRepoTree(QTreeWidget):
    '''Widget managing a list of repositories'''

    rmbMenu: QMenu
    rmbTgitMenu: QMenu

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)

        # to allow right-mouse-button click signal
        self.setContextMenuPolicy(Qt.CustomContextMenu)

        f = self.header().font()
        f.setBold(True)
        # the column for the refresh icon
        self.headerItem().setText(0, '')
        self.header().setFont(f)
        self.sortByColumn(mg_const.COL_REPO_NAME, Qt.AscendingOrder)
        self.setColumnWidth(mg_const.COL_UPDATE, 40)

        if self.columnCount() < mg_const.COL_NB:
            self.setColumnCount(mg_const.COL_NB)
            self.setHeaderLabels(mg_const.COL_TITLES)
            self.setSelectionMode(QAbstractItemView.ExtendedSelection)
            self.setAllColumnsShowFocus(True)
            self.header().setSortIndicatorShown(True)

        # calling slot to adjust the status of the column
        self.slotViewColSha1Changed(mgc.get_config_instance().get(mgc.CONFIG_VIEW_COL_SHA1, True))
        self.slotViewColUrlChanged (mgc.get_config_instance().get(mgc.CONFIG_VIEW_COL_URL, True))

        self.gitExecWindow: Optional[MgExecWindow] = None

        self.mgActions = MgActions(self)
        self.rmbMenu = QMenu(self)
        self.mgActions.setupMenuView(self.rmbMenu)
        self.rmbGitMenu = self.rmbMenu.addMenu("Git")
        self.rmbGitMenu.setIcon(QIcon(":/img/git_black.png"))
        self.mgActions.setupMenuGit(self.rmbGitMenu)
        self.mgActions.setupMenuGitPrograms(self.rmbMenu, includeConfigureGitPrograms=False)
        self.setupConnections()
        pluginMgrInstance.setupRepoRmbMenu(self)


    def clear(self) -> None:
        '''Called when clearing the list of repoItem and also of removing all repoInfo items.'''
        # we want to disconnect all signals emitted from repoInfo to our items, to avoid C++ object already
        # deleted being called by Python objects still present
        for repoItem in self.allRepoItems():
            # disconnect all signals, what so ever
            # it does not work on PySide2
            # QObject.disconnect(cast(QObject, repoItem.repoInfo))
            pass

        super().clear()


    def slotViewColSha1Changed(self, showColSha1: bool) -> None:
        '''Called when menu item View->Columns->Sha1 changes value'''
        self.setColumnHidden(mg_const.COL_SHA1, not showColSha1)
        if showColSha1:
            # ensure that sha1 information is fetched from all repos
            for repoItem in self.allRepoItems():
                repoItem.repoInfo.ensure_sha1(repoItem.cbSha1available)


    def slotViewColUrlChanged(self, showColUrl: bool) -> None:
        '''Called when menu item View->Columns->Url changes value'''
        self.setColumnHidden(mg_const.COL_URL, not showColUrl)
        if showColUrl:
            # ensure that URL information is fetched from all repos
            for repoItem in self.allRepoItems():
                repoItem.repoInfo.ensure_url(repoItem.cbUrlAvailable)


    def availableScmUpdated(self) -> None:
        '''Called when the configuration for available SCM has changed, to update our RMB menu'''
        self.mgActions.enableAvailableScm()


##########################################################################
#
#       Menu and actions setup
#
##########################################################################

    def setupConnections(self) -> None:
        '''Install all action connections'''

        self.customContextMenuRequested.connect(self.slotContextMenuRequested)
        self.itemActivated.connect(self.slotItemActivated)

        # menu Edit
        self.mgActions.actionRefreshSelected.triggered.connect(self.slotRefreshSelected)
        self.mgActions.actionShowInExplorer.triggered.connect(self.slotShowInExplorer)
        self.mgActions.menuCopy.triggered.connect(self.slotMenuCopyAction)

        # Menu Git
        self.mgActions.actionGitProperties.triggered.connect(self.slotGitProperties)
        self.mgActions.actionGitCommit.triggered.connect(self.slotGitCommit)
        self.mgActions.actionGitRevert.triggered.connect(self.slotGitRevert)
        self.mgActions.actionGitPush.triggered.connect(self.slotGitPush)
        self.mgActions.actionGitPushTag.triggered.connect(self.slotGitPushTag)
        self.mgActions.actionGitPull.triggered.connect(self.slotGitPull)
        self.mgActions.actionGitFetch.triggered.connect(self.slotGitFetch)
        self.mgActions.actionGitTag.triggered.connect(self.slotGitTag)
        self.mgActions.actionGitCreateBranch.triggered.connect(self.slotGitCreateBranch)
        self.mgActions.actionGitSwitchBranch.triggered.connect(self.slotGitSwitchBranch)
        self.mgActions.actionGitCheckoutTag.triggered.connect(self.slotGitCheckoutTag)
        self.mgActions.actionGitDeleteBranch.triggered.connect(self.slotGitDeleteBranch)
        self.mgActions.actionGitBash.triggered.connect(self.slotGitBash)
        self.mgActions.actionGitRunCommand.triggered.connect(self.slotGitRunCommand)

        # Menu Git programs
        self.mgActions.actionSourceTree.triggered.connect(self.slotSourcetree)
        self.mgActions.actionSublimeMerge.triggered.connect(self.slotSublimemerge)
        self.mgActions.actionGitGui.triggered.connect(self.slotGitGui)
        self.mgActions.actionGitK.triggered.connect(self.slotGitK)

        self.mgActions.actionTGitShowLog  .triggered.connect(self.slotTGitShowLog)
        self.mgActions.actionTGitCommit   .triggered.connect(self.slotTGitCommit)
        self.mgActions.actionTGitDiff     .triggered.connect(self.slotTGitDiff)
        self.mgActions.actionTGitRevert   .triggered.connect(self.slotTGitRevert)
        self.mgActions.actionTGitPush     .triggered.connect(self.slotTGitPush)
        self.mgActions.actionTGitPull     .triggered.connect(self.slotTGitPull)
        self.mgActions.actionTGitFetch    .triggered.connect(self.slotTGitFetch)
        self.mgActions.actionTGitTag      .triggered.connect(self.slotTGitTag)
        self.mgActions.actionTGitSwitch   .triggered.connect(self.slotTGitSwitch)
        self.mgActions.actionTGitBranch   .triggered.connect(self.slotTGitBranch)

        self.mgActions.menuCopy.aboutToShow.connect(self.slotMenuCopyAboutToShow)
        self.mgActions.menuCopy.aboutToHide.connect(self.slotMenuCopyAboutToHide)


    def slotContextMenuRequested(self, p: QPoint) -> None:
        '''Capture the right-button mouse release action to display a menu'''
        self.rmbMenu.exec_(self.viewport().mapToGlobal(p)) # qme.globalPos())


##########################################################################
#
#       Item and repo management
#
##########################################################################

    def deleteRepos(self, deletedRepo: List[MgRepoInfo]) -> None:
        '''Remove all the items related to the listed repositories'''
        for repoInfo in deletedRepo:
            for itemIdx in range(self.topLevelItemCount()):
                if self.topLevelItem(itemIdx).text(mg_const.COL_REPO_NAME) == repoInfo.name:     # type: ignore # topLevelItem(itemIdx) exists and is not None
                    self.takeTopLevelItem(itemIdx)
                    break  # we only have one such item per repo, no need to continue searching
            else:
                warn('Item not found for deleted repo: %s' % repoInfo.name)

        # show visually that the repo were deleted
        QApplication.processEvents()


    def addRepos(self, repoInfoList: List[MgRepoInfo]) -> List[MgRepoTreeItem]:
        '''Add the listed repos as items to this repoTree.
        Return the list of MgRepoTreeItem created'''
        noSelectedItem = (self.topLevelItemCount() == 0)

        items: List[MgRepoTreeItem] = []

        # first step, we fill the list of repo
        for repoInfo in repoInfoList:
            item = MgRepoTreeItem(repoInfo, self)
            items.append(item)
            if noSelectedItem:
                self.setCurrentItem(item)
                noSelectedItem = False

            self.autoAdjustColumnSize()

        return items


    def autoAdjustColumnSize(self) -> None:
        '''Adjust automatically the column size to the largest item'''
        for i in range(self.columnCount()):
            self.resizeColumnToContents(i)
        QApplication.processEvents()


    def selectedRepoItems(self) -> List[MgRepoTreeItem]:
        '''Return the list of selected MgRepoTreeItem'''
        items = cast(List[MgRepoTreeItem], list(super().selectedItems()))
        return items


    def selectedRepos(self) -> List[MgRepoInfo]:
        '''Return the list of selected Repositories'''
        repos = [item.repoInfo for item in self.selectedRepoItems()]
        return repos


    def allRepos(self) -> List[MgRepoInfo]:
        '''Return the list of all Repositories'''
        repos = [item.repoInfo for item in self.allRepoItems()]
        return repos


    def allRepoItems(self) -> List[MgRepoTreeItem]:
        '''Return the list of all MgRepoTreeItem of the widget'''
        items = []
        for idx in range(self.topLevelItemCount()):
            items.append(cast(MgRepoTreeItem, self.topLevelItem(idx)))
        return items

##########################################################################
#
#       Actions on selected items
#
##########################################################################

    def getGitExecWindow(self) -> MgExecWindow:
        '''Get the running git execution window if any, or create it if necessary'''
        if self.gitExecWindow is None:
            self.gitExecWindow = MgExecWindow(self)
            self.gitExecWindow.finished.connect(self.clearGitExecWindow)
        return self.gitExecWindow


    def clearGitExecWindow(self) -> None:
        '''Called when execution is done, to clear the variable.

        That way, new tasks will use a new instance.'''
        # this prevents a bug where the window was hidden but still visible when hovering
        # over the taskbar window of MultiGit
        del self.gitExecWindow
        self.gitExecWindow = None


    def confirmIfNoSelectedItems(self) -> bool:
        '''Display a warning of no items are selected and return False. Return True if everything is OK'''
        items = self.selectedRepoItems()
        if len(items) == 0:
            QMessageBox.warning(self, 'No repository selected',
                                'Warning: you must select at least one repository to run this command!')
            return False
        return True


    def confirmIfTooManySelectedItems(self, dialogName: str) -> bool:
        '''If too many items are selected, ask the user to confirm the action.
        Returns True if the user confirmed, else False.
        '''
        items = self.selectedRepoItems()
        if len(items) <= 3:
            return True
        
        buttonSelected = QMessageBox.warning(self, 'Many repositories selected',
                                             'You are about to launch %d %s\nDo you want to continue ?' % (len(items), dialogName),
                                             QMessageBox.Cancel | QMessageBox.Ok, QMessageBox.Cancel)
        return not (buttonSelected != QMessageBox.Ok)


    def confirmIfNoOrTooManySelectedItems(self, dialogName: str) -> bool:
        '''Convenience method grouping confirmIfNoSelectedItems() and confirmIfTooManySelectedItems().
        Returns False if we should abort processing
        '''
        return self.confirmIfTooManySelectedItems(dialogName) and self.confirmIfNoSelectedItems()


    def slotItemActivated(self, _item: QTreeWidgetItem, _col: int) -> None:
        '''Called when user double-clicked or presses enter on an item. Call user-defined action'''
        dbg('slotItemActivated()')

        # Make sure you don't forget any action!
        actionToPerform = {
            mg_const.DBC_GITCOMMIT: self.slotGitCommit,
            mg_const.DBC_GITCREATEBRANCH     : self.slotGitCreateBranch,
            mg_const.DBC_GITSWITCHBRANCH     : self.slotGitSwitchBranch,
            mg_const.DBC_GITPUSH             : self.slotGitPush,
            mg_const.DBC_GITPULL             : self.slotGitPull,
            mg_const.DBC_GITFETCH            : self.slotGitFetch,
            mg_const.DBC_REPOSITORYPROPERTIES: self.slotGitProperties,
            mg_const.DBC_SHOWINEXPLORER      : self.slotShowInExplorer,

            mg_const.DBC_DONOTHING           : (lambda : None),

            mg_const.DBC_TORTOISEGITSHOWLOG  : self.slotTGitShowLog,
            mg_const.DBC_TORTOISEGITSWITCH   : self.slotTGitSwitch,
            mg_const.DBC_TORTOISEGITBRANCH   : self.slotTGitBranch,
            mg_const.DBC_TORTOISEGITCOMMIT   : self.slotTGitCommit,
            mg_const.DBC_TORTOISEGITDIFF     : self.slotTGitDiff,
            mg_const.DBC_TORTOISEGITPUSH     : self.slotTGitPush,
            mg_const.DBC_TORTOISEGITPULL     : self.slotTGitPull,
            mg_const.DBC_TORTOISEGITFETCH    : self.slotTGitFetch,

            mg_const.DBC_RUNSOURCETREE       : self.slotSourcetree,
            mg_const.DBC_RUNSUBLIMEMERGE     : self.slotSublimemerge,
            mg_const.DBC_RUNGITGUI           : self.slotGitGui,
            mg_const.DBC_RUNGITK             : self.slotGitK,
            mg_const.DBC_RUNGITBASH          : self.slotGitBash,

        } # type: Dict[str, Callable[[], None]]

        if mgc.get_config_instance()[mgc.CONFIG_DOUBLE_CLICK_ACTION] in actionToPerform:
            actionToPerform[mgc.get_config_instance()[mgc.CONFIG_DOUBLE_CLICK_ACTION]]()
        else:
            button = QMessageBox.question(self, "Action for double-click", "Action for double-click is not yet defined.\nDo you want to open the settings dialog to define it ?")
            if button == QMessageBox.Yes:
                runDialogEditSettings(self, tabPage=0)


    def slotRefreshSelected(self) -> None:
        dbg('slotRefreshSelected()')
        if not self.confirmIfNoSelectedItems():
            return
        self.doRefreshItems(self.selectedRepoItems())


    def doRefreshItems(self, items: List[MgRepoTreeItem]) -> None:
        for it in items:
            # clear items before updating them, this is better visually
            it.markItemInProgress()

        for it in items:
            it.repoInfo.refresh()


    def slotMenuCopyAboutToShow(self) -> None:
        '''Called when user clicked on the copy submenu. Fills the items of the menu'''
        items = self.selectedRepoItems()
        if len(items) == 0:
            return

        repoInfo = items[0].repoInfo
        self.mgActions.setupDynamicMenuCopy(repoInfo)


    def slotMenuCopyAboutToHide(self) -> None:
        '''Called when Copy menu is going to hide. We disable the connections created
        by the Copy menu to capture the repo missing info'''

        # connections are used to update the menu after is has been shown, when
        # information is missing. We must disconnect everything when the menu
        # goes to hide
        self.mgActions.clearMenuCopyConnections()


    def slotMenuCopyAction(self, action: QAction) -> None:
        '''An item in the copy menu has been selected'''
        text = action.text()
        dbg(f'slotMenuCopyAction() - setting clipboard to "{text}"')
        QApplication.clipboard().setText(text)


    def slotShowInExplorer(self) -> None:
        '''Open one explorer window for each git repo'''
        dbg('slotShowInExplorer()')
        if not self.confirmIfNoOrTooManySelectedItems('Explorer'):
            return

        if not ExecExplorer.checkFound():
            return

        for it in self.selectedRepoItems():
            ExecExplorer.exec_non_blocking([it.repoInfo.fullpath])


    ##########################################################################
    #
    #       Git actions
    #
    ##########################################################################

    def runGitCommand(self, desc: str, cmd: List[str]) -> None:
        '''Run a git command on all selected repos

        desc: a description of the command to run, used as title of the execution dialog
        cmd: a list of string of a git command to run, without the git itself

        Dialog rejecting command if no item is selected should be shown prior to calling this.
        '''
        items = self.selectedRepoItems()
        if len(items) == 0:
            QMessageBox.warning(self, 'No repository selected',
                                'Warning: you must select at least one repository to run this command!')
            return

        # collect all repos to act upon
        repos = [ item.repoInfo for item in items ]

        # show window for executing git
        self.getGitExecWindow().execOneGitCommand(desc, cmd, repos)


    def slotGitProperties(self) -> None:
        dbg('slotGitProperties()')
        if not self.confirmIfNoSelectedItems():
            return
        runDialogGitProperties(self, self.selectedRepoItems()[0].repoInfo)


    def slotGitCommit(self) -> None:
        '''Commit on all listed repos'''
        dbg('slotGitCommit()')
        if not self.confirmIfNoSelectedItems():
            return

        # local import to avoid circular dependency
        from src.mg_dialog_git_commit import runDialogGitCommit
        runDialogGitCommit(self, self.selectedRepos(), self.allRepos())


    def slotGitRevert(self) -> None:
        '''Run git Revert on all listed repos'''
        dbg('slotGitRevert()')
        if not self.confirmIfNoSelectedItems():
            return

        # local import to avoid circular dependency
        from src.mg_dialog_git_revert import runDialogGitRevert
        runDialogGitRevert(self, self.selectedRepos(), self.allRepos())


    def slotGitPush(self) -> None:
        '''Run git Push on all listed repos'''
        dbg('slotGitPush()')
        if not self.confirmIfNoSelectedItems():
            return
        items = self.selectedRepoItems()

        desc = 'Pushing current branch'

        repoAndListOfGitCmd = []
        # collect all repos to act upon
        repos = [item.repoInfo for item in items]
        for repoInfo in repos:
            cmd = ['push', '-u', '--progress', '--verbose', 'origin', '{0}'.format(repoInfo.branch)]
            repoAndListOfGitCmd.append(('Push branch %s' % repoInfo.branch,
                                        repoInfo,
                                        [cmd],
                                        ))

        # show window for executing git
        self.getGitExecWindow().execEachRepoWithHisSeqOfGitCommand(desc, repoAndListOfGitCmd)


    def slotGitPushTag(self) -> None:
        '''Choose tags to push for all repos'''
        dbg('slotGitPushTag()')
        if not self.confirmIfNoSelectedItems():
            return

        from src.mg_dialog_git_push_tag import runDialogGitPushTag
        runDialogGitPushTag(self, self.selectedRepos(), self.allRepos())



    def slotGitPull(self) -> None:
        '''Run git Pull on all listed repos'''
        dbg('slotGitPull()')
        # checking for empty selected repos is done inside runGitCommand()
        self.runGitCommand('Pulling repositories', ['pull', '--verbose'])


    def slotGitFetch(self) -> None:
        '''Run git fetch on all listed repos'''
        dbg('slotGitFetch()')
        # checking for empty selected repos is done inside runGitCommand()
        self.runGitCommand('Fetching repositories', ['fetch', '--verbose', '--prune'])


    def slotGitTag(self) -> None:
        '''Run git Tag on all listed repos'''
        dbg('slotGitTag()')
        if not self.confirmIfNoSelectedItems():
            return

        # local import to prevent circular dependencies
        from src.mg_dialog_git_tag import runDialogGitTag
        runDialogGitTag(self, self.selectedRepos(), self.allRepos())


    def slotGitCreateBranch(self) -> None:
        '''Create a branch on all listed repos'''
        dbg('slotGitCreateBranch()')
        if not self.confirmIfNoSelectedItems():
            return

        # local import to prevent circular dependencies
        from src.mg_dialog_git_create_branch import runDialogGitCreateBranch
        runDialogGitCreateBranch(self, self.selectedRepos(), self.allRepos())


    def slotGitSwitchBranch(self) -> None:
        '''Switch to a branch on all listed repos'''
        dbg('slotGitSwitchBranch()')
        # local import to prevent circular dependencies
        from src.mg_dialog_git_switch_delete_branch import runDialogGitSwitchDelete, DeleteOrSwitch
        runDialogGitSwitchDelete(self, DeleteOrSwitch.SWITCH_BRANCH, self.selectedRepos(), self.allRepos())


    def slotGitCheckoutTag(self) -> None:
        '''Switch to a tag on all listed repos'''
        dbg('slotGitCheckoutTag()')
        # local import to prevent circular dependencies
        from src.mg_dialog_git_switch_delete_branch import runDialogGitSwitchDelete, DeleteOrSwitch
        runDialogGitSwitchDelete(self, DeleteOrSwitch.CHECKOUT_TAG, self.selectedRepos(), self.allRepos())


    def slotGitDeleteBranch(self) -> None:
        '''Delete branch on all listed repos'''
        dbg('slotGitDeleteBranch()')
        # local import to prevent circular dependencies
        from src.mg_dialog_git_switch_delete_branch import runDialogGitSwitchDelete, DeleteOrSwitch
        runDialogGitSwitchDelete(self, DeleteOrSwitch.DELETE, self.selectedRepos(), self.allRepos())


    def slotGitBash(self) -> None:
        '''Run git-bash on the current repos'''
        dbg('slotGitBash')
        if not self.confirmIfNoOrTooManySelectedItems('Git bash'):
            return

        try:
            for item in self.selectedRepoItems():
                repo = item.repoInfo
                # after git bash, we also want to refresh the URL
                ExecGitBash.exec_non_blocking([], workdir=repo.fullpath, callback=repo.deepRefresh)
        except FileNotFoundError:
            # Git Bash was not located
            QMessageBox.warning(self, 'Unable to execute Git Bash', 'Warning: could not locate the git-bash.exe program.\n' +
                                'Can not execute any Git Bash command..\nPlease configure the location in the settings dialog.\n')
            self.mgActions.actionGitBash.setEnabled(False)


    def slotGitRunCommand(self) -> None:
        '''Run a custom git command'''
        dbg('slotGitRunCommand')
        if not self.confirmIfNoSelectedItems():
            return

        # local import to prevent circular dependencies
        from src.mg_dialog_git_run_cmd import runDialogGitCommand
        runDialogGitCommand(self, self.selectedRepos(), self.allRepos())



    ##########################################################################
    #
    #       TortoiseGit actions
    #
    ##########################################################################

    def tortoiseGitCommandOnSelection(self, dialogName: str, cmd: str) -> None:
        '''Verify the condition for running a TortoiseGit dialog.

        Check that one repository is selected, if not warn the user.
        If too many repositories are selected, ask the user to validate.

        dialogName: the name of the dialog to subtitute in the error message
        cmd: the tortoise git command to run
        '''
        if not self.confirmIfNoOrTooManySelectedItems(dialogName):
            return

        self.tortoiseGitCommandOnItems(cmd, self.selectedRepoItems())


    def tortoiseGitCommandOnItems(self, cmd: str, items: List[MgRepoTreeItem]) -> None:
        '''Run a Tortoise Git command on all repos passed in argument'''
        try:
            for item in items:
                repo = item.repoInfo
                # put the path parameter in a separate argument, see: https://gitlab.com/tortoisegit/tortoisegit/-/issues/3518
                ExecTortoiseGit.exec_non_blocking(['/path', repo.fullpath, '/command:%s' % cmd], callback=repo.refresh)
        except FileNotFoundError:
            # TortoiseGit was not located
            QMessageBox.warning(self, 'Unable to execute TortoiseGit', 'Warning: could not locate the TortoiseGitProc.exe program.\n' +
                                'Can not execute any TortoiseGit command..\nPlease configure the location in the settings dialog.\n')


    def slotTGitShowLog(self) -> None:
        dbg('slotTGitShowLog()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Show Log', 'log')


    def slotTGitCommit(self) -> None:
        dbg('slotTGitCommit()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Commit', 'commit')


    def slotTGitDiff(self) -> None:
        dbg('slotTGitDiff()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Diff', 'diff')


    def slotTGitRevert(self) -> None:
        dbg('slotTGitRevert()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Revert', 'revert')


    def slotTGitPull(self) -> None:
        dbg('slotTGitPull()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Pull', 'pull')


    def slotTGitPush(self) -> None:
        dbg('slotTGitPush()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Push', 'push')


    def slotTGitFetch(self) -> None:
        dbg('slotTGitFetch()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Fetch', 'fetch')


    def slotTGitTag(self) -> None:
        dbg('slotTGitTag()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Tag', 'tag')


    def slotTGitSwitch(self) -> None:
        dbg('slotTGitSwitch()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Switch', 'switch')


    def slotTGitBranch(self) -> None:
        dbg('slotTGitBranch()')
        self.tortoiseGitCommandOnSelection('TortoiseGit Branch', 'branch')


    ##########################################################################
    #
    #       Other actions
    #
    ##########################################################################


    def slotSourcetree(self) -> None:
        '''Run sourcetree on the current repos'''
        dbg('slotSourcetree')
        if not self.confirmIfNoOrTooManySelectedItems('SourceTree'):
            return

        try:
            for item in self.selectedRepoItems():
                repo = item.repoInfo
                # run sourcetree
                ExecSourceTree.exec_non_blocking(['-t', str(repo.fullpath)], callback=repo.refresh)
        except FileNotFoundError:
            # TortoiseGit was not located
            QMessageBox.warning(self, 'Unable to execute SourceTree', 'Warning: could not locate the SourceTree.exe program.\n' +
                                'Can not execute any SourceTree command..\nPlease configure the location in the settings dialog.\n')
            self.mgActions.actionSourceTree.setEnabled(False)


    def slotGitGui(self) -> None:
        '''Run git Gui on the current repos'''
        dbg('slotGitGui')
        if not self.confirmIfNoOrTooManySelectedItems('Git GUI'):
            return

        try:
            for item in self.selectedRepoItems():
                repo = item.repoInfo
                # allow errors needed because git-gui never returns 0
                ExecGitGui.exec_non_blocking([], workdir=str(repo.fullpath), allow_errors=True, callback=repo.refresh)
        except FileNotFoundError:
            QMessageBox.warning(self, 'Unable to execute Git GUI', 'Warning: could not locate the git-gui.exe program.\n' +
                                'Can not execute any git-gui command..\nPlease configure the location in the settings dialog.\n')
            self.mgActions.actionGitGui.setEnabled(False)


    def slotSublimemerge(self) -> None:
        '''Run SublimeMewrge on the current repos'''
        dbg('slotSublimemerge')
        if not self.confirmIfNoOrTooManySelectedItems('Sublime Merge'):
            return

        try:
            for item in self.selectedRepoItems():
                repo = item.repoInfo
                ExecSublimeMerge.exec_non_blocking([str(repo.fullpath)], callback=repo.refresh)
        except FileNotFoundError:
            # TortoiseGit was not located
            QMessageBox.warning(self, 'Unable to execute SublimeMerge', 'Warning: could not locate the SublimeMerge.exe program.\n' +
                                'Can not execute any SublimeMerge command..\nPlease configure the location in the settings dialog.\n')
            self.mgActions.actionSublimeMerge.setEnabled(False)


    def slotGitK(self) -> None:
        '''Run GitK on the current repos'''
        dbg('slotGitK')
        if not self.confirmIfNoOrTooManySelectedItems('GitK'):
            return

        try:
            for item in self.selectedRepoItems():
                repo = item.repoInfo
                # ExecGitK.exec_non_blocking([], workdir=str(repo.fullpath), callback=repo.refresh)
                ExecGitK.exec_non_blocking([], workdir=str(repo.fullpath), allow_errors=True,callback=repo.refresh)
        except FileNotFoundError:
            QMessageBox.warning(self, 'Unable to execute GitK', 'Warning: could not locate the gitk.exe program.\n' +
                                'Can not execute any GitK command..\nPlease configure the location in the settings dialog.\n')
            self.mgActions.actionSublimeMerge.setEnabled(False)





#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List, Tuple

import unittest

from src.mg_dialog_clone_from_mgit import addPreconditionToEnsureCloneOrderLogic, \
    build_taskgroup_dep_graph, addPreconditionsToTaskGroup
from src.mg_exec_task_item import MgExecTaskGroup
from src.mg_repo_info import MgRepoInfo


class TestCloneGraphOrdering(unittest.TestCase):

    def buildTaskGroups(self) -> Tuple[List[str], List[MgExecTaskGroup]]:
        repoNames = [
            'dev',
            'dev/subdev1',
            'dev/subdev1/subdev1_sub1',
            'dev/subdev1/subdev1_sub2',
            'dev/subdev2',
            'dev/subdev2/toto/subdev2_sub1',
            'test',
            'test/extern/subtest1',
            'doc/whats/up',
        ]


        taskGroups: List[MgExecTaskGroup] = [
            MgExecTaskGroup(name, MgRepoInfo(name, name))
            for name in repoNames
        ]

        return repoNames, taskGroups



    def test_cloneGraphOrdering(self):
        repoNames, taskGroups = self.buildTaskGroups()
        taskGroups = addPreconditionToEnsureCloneOrderLogic(taskGroups)

        # check that ordering is the same
        self.assertEqual([tg.desc for tg in taskGroups], repoNames)

    def test_cloneGraphStructureFlexibleSplitter(self):
        repoNames = [
            'titi/toto',
            'tata\\tutu',
        ]

        taskGroups: List[MgExecTaskGroup] = [
            MgExecTaskGroup(name, MgRepoInfo(name, name))
            for name in repoNames
        ]
        topNode = build_taskgroup_dep_graph(taskGroups)

        self.assertEqual(topNode.children[0].name, 'titi')
        self.assertEqual(topNode.children[0].children[0].name, 'toto')

        self.assertEqual(topNode.children[1].name, 'tata')
        self.assertEqual(topNode.children[1].children[0].name, 'tutu')


    def test_cloneGraphStructure(self):
        repoNames, taskGroups = self.buildTaskGroups()
        topNode = build_taskgroup_dep_graph(taskGroups)

        self.assertEqual(topNode.children[0].name, 'dev')
        self.assertEqual(topNode.children[0].taskGroup.desc, 'dev')
        self.assertEqual(topNode.children[1].name, 'test')
        self.assertEqual(topNode.children[1].taskGroup.desc, 'test')
        self.assertEqual(topNode.children[2].name, 'doc')
        self.assertEqual(topNode.children[2].taskGroup, None)
        self.assertEqual(len(topNode.children), 3)

        node = topNode.children[0]
        self.assertEqual(node.children[0].name, 'subdev1')
        self.assertEqual(node.children[0].taskGroup.desc, 'dev/subdev1')
        self.assertEqual(node.children[1].name, 'subdev2')
        self.assertEqual(node.children[1].taskGroup.desc, 'dev/subdev2')
        self.assertEqual(len(node.children), 2)

        node = topNode.children[0].children[0]
        self.assertEqual(node.children[0].name, 'subdev1_sub1')
        self.assertEqual(node.children[0].taskGroup.desc, 'dev/subdev1/subdev1_sub1')
        self.assertEqual(node.children[1].name, 'subdev1_sub2')
        self.assertEqual(node.children[1].taskGroup.desc, 'dev/subdev1/subdev1_sub2')
        self.assertEqual(len(node.children), 2)

        node = topNode.children[0].children[1]
        self.assertEqual(node.children[0].name, 'toto')
        self.assertEqual(node.children[0].taskGroup, None)
        self.assertEqual(len(node.children), 1)

        node = topNode.children[0].children[1].children[0]
        self.assertEqual(node.children[0].name, 'subdev2_sub1')
        self.assertEqual(node.children[0].taskGroup.desc, 'dev/subdev2/toto/subdev2_sub1')
        self.assertEqual(len(node.children), 1)

        node = topNode.children[1]
        self.assertEqual(node.children[0].name, 'extern')
        self.assertEqual(node.children[0].taskGroup, None)
        self.assertEqual(len(node.children), 1)

        node = topNode.children[1].children[0]
        self.assertEqual(node.children[0].name, 'subtest1')
        self.assertEqual(node.children[0].taskGroup.desc, 'test/extern/subtest1')
        self.assertEqual(len(node.children), 1)

        node = topNode.children[2]
        self.assertEqual(node.children[0].name, 'whats')
        self.assertEqual(node.children[0].taskGroup, None)
        self.assertEqual(len(node.children), 1)

        node = topNode.children[2].children[0]
        self.assertEqual(node.children[0].name, 'up')
        self.assertEqual(node.children[0].taskGroup.desc, 'doc/whats/up')
        self.assertEqual(len(node.children), 1)


    def test_cloneGraphPreconditions(self):
        repoNames, taskGroups = self.buildTaskGroups()
        topNode = build_taskgroup_dep_graph(taskGroups)
        addPreconditionsToTaskGroup(topNode)

        # check each precondition

        self.assertEqual(taskGroups[0].desc, 'dev')
        self.assertEqual(taskGroups[0].pre_condition, None)

        self.assertEqual(taskGroups[1].desc, 'dev/subdev1')
        self.assertEqual(taskGroups[1].pre_condition.task_group, taskGroups[0])

        self.assertEqual(taskGroups[2].desc, 'dev/subdev1/subdev1_sub1')
        self.assertEqual(taskGroups[2].pre_condition.task_group, taskGroups[1])

        self.assertEqual(taskGroups[3].desc, 'dev/subdev1/subdev1_sub2')
        self.assertEqual(taskGroups[3].pre_condition.task_group, taskGroups[1])

        self.assertEqual(taskGroups[4].desc, 'dev/subdev2')
        self.assertEqual(taskGroups[4].pre_condition.task_group, taskGroups[0])

        self.assertEqual(taskGroups[5].desc, 'dev/subdev2/toto/subdev2_sub1')
        self.assertEqual(taskGroups[5].pre_condition.task_group, taskGroups[4])

        self.assertEqual(taskGroups[6].desc, 'test')
        self.assertEqual(taskGroups[6].pre_condition, None)

        self.assertEqual(taskGroups[7].desc, 'test/extern/subtest1')
        self.assertEqual(taskGroups[7].pre_condition.task_group, taskGroups[6])

        self.assertEqual(taskGroups[8].desc, 'doc/whats/up')
        self.assertEqual(taskGroups[8].pre_condition, None)

        # TaskGroups:
        # - 'dev'
        # - 'dev/subdev1'
        # - 'dev/subdev1/subdev1_sub1'
        # - 'dev/subdev1/subdev1_sub2'
        # - 'dev/subdev2'
        # - 'dev/subdev2/toto/subdev2_sub1'
        # - 'test'
        # - 'test/extern/subtest1'
        # - 'doc/whats/up'





#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
import sys
from typing import TYPE_CHECKING, List, Optional
import logging, pathlib, json, os, subprocess, time

from PySide2.QtGui import QRegExpValidator, QColor
from PySide2.QtWidgets import QDialog, QWidget, QFileDialog, QMessageBox, QTreeWidgetItem, QPushButton, QDialogButtonBox
from PySide2.QtCore import Qt, QTimer, QRegExp

if TYPE_CHECKING:
    from src.mg_window import MgMainWindow
from src.gui.ui_clone_from_mgit import Ui_CloneFromMgitFile
from src.mg_json_mgit_parser import ProjectStructure
from src.mg_repo_info import MgRepoInfo
import src.mg_config as mgc
from src.mg_exec_window import MgExecWindow
from src.mg_exec_task_item import MgExecTaskGroup, after_other_taskgroup_is_started_and_dir_exists
from src.mg_dialog_utils import reBranchTagValues
from src.mg_utils import set_username_on_git_url, deleteDirList
from src.mg_plugin_mgr import pluginMgrInstance

logger = logging.getLogger('mg_dialog_clone_from_mgit')
dbg = logger.debug
warn = logger.warning


# Cases:
# 1. projConfFile and projDestDir are valid
# 2. update projConfFile
#       2.a -> parse the file -> valid file -> update full GUI
#       2.b -> parse the file -> invalid file -> report error, clear the GUI
# 3. update projDestDir
#       3.a update proj repos destination -> update the display of repo items
#       3.b json file was invalid -> do nothing



class MgDialogCloneFromMgitFile(QDialog):

    def __init__(self, window: QWidget) -> None:
        super().__init__(window)
        # noinspection PyTypeChecker
        self.setWindowFlags( self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self.ui = Ui_CloneFromMgitFile()
        self.ui.setupUi(self)
        self.ui.labelProjDesc.setVisible(False)
        self.ui.textEditProjectDesc.setVisible(False)
        self.cloneButton = QPushButton('Clone', self)
        self.ui.buttonBox.addButton(self.cloneButton, QDialogButtonBox.AcceptRole)

        self.userFinishedTypingMgitFileTimer = QTimer(self)
        self.userFinishedTypingMgitFileTimer.setSingleShot(True)
        self.userFinishedTypingMgitFileTimer.setInterval(1000)
        self.userFinishedTypingMgitFileTimer.timeout.connect(lambda: self.propagateMgitFileUpdated(False, True))
        self.ui.lineEditMgitFile.textEdited.connect(self.slotMgitFileTimer)

        self.userFinishedTypingDestDirTimer = QTimer(self)
        self.userFinishedTypingDestDirTimer.setSingleShot(True)
        self.userFinishedTypingDestDirTimer.setInterval(1000)
        self.userFinishedTypingDestDirTimer.timeout.connect(self.updateDestDir)
        self.ui.lineEditDestDir.textEdited.connect(self.slotDestDirTimer)

        self.ui.radioDoNotAlterUrl.clicked.connect(self.slotUsernameUpdated)
        self.ui.radioForceUsername.clicked.connect(self.slotUsernameUpdated)
        self.ui.radioStripUsername.clicked.connect(self.slotUsernameUpdated)

        self.userFinishedTypingUsernameTimer = QTimer(self)
        self.userFinishedTypingUsernameTimer.setSingleShot(True)
        self.userFinishedTypingUsernameTimer.setInterval(1000)
        self.userFinishedTypingUsernameTimer.timeout.connect(self.slotUsernameUpdated)
        self.ui.lineEditUsername.textEdited.connect(self.slotUsernameTimer)

        self.proj = ProjectStructure()
        self.proj_copy: Optional[ProjectStructure] = None

        validator = QRegExpValidator(QRegExp(reBranchTagValues))
        self.ui.lineEditUsername.setValidator(validator)

        # set last values
        self.ui.historyButtonMgitFile.fillHistory(mgc.get_config_instance().lruAsList(mgc.CONFIG_LAST_MGIT_FILE))
        self.ui.lineEditMgitFile.setText('')
        self.ui.historyButtonDestDir.fillHistory(mgc.get_config_instance().lruAsList(mgc.CONFIG_LAST_PROJECT_DIR))
        self.ui.lineEditDestDir.setText('')
        self.ui.historyButtonUsername.fillHistory(mgc.get_config_instance().lruAsList(mgc.CONFIG_CLONE_USERNAME))

        # configure QtreeWidget to display list of repositories
        f = self.ui.treeWidgetRepoList.header().font()
        f.setBold(True)
        self.ui.treeWidgetRepoList.header().setFont(f)
        self.ui.treeWidgetRepoList.sortByColumn(0, Qt.AscendingOrder)

        # if both items are filled and exist, updates repos

        # connect buttons to their slot
        self.ui.pushButtonChooseMgitFile.clicked.connect(self.slotChooseMgitFile)
        self.ui.pushButtonChooseDestDir.clicked.connect(self.slotChooseDestDir)
        self.ui.historyButtonMgitFile.historyItemTriggered.connect(self.slotMgitFileHistoryTriggered)
        self.ui.historyButtonDestDir.historyItemTriggered.connect(self.slotDestDirHistoryTriggered)
        self.ui.historyButtonUsername.historyItemTriggered.connect(self.slotUsernameHistoryTriggered)

        pluginMgrInstance.setupCloneDialog(self)

        QTimer.singleShot(500, lambda: self.propagateMgitFileUpdated(False, True))


    def slotMgitFileTimer(self, _text: str) -> None:
        '''Called when user modifies the text'''
        # we wait for 0.5 s then refresh the analysis of the file
        self.userFinishedTypingMgitFileTimer.start()


    def slotDestDirTimer(self, _text: str) -> None:
        '''Called when user modifies the text'''
        # we wait for 0.5 s then refresh the analysis of the file
        self.userFinishedTypingDestDirTimer.start()


    def slotUsernameTimer(self, _text: str) -> None:
        '''Called when user modifies the text of the username'''
        # we wait for 0.5 s then refresh the analysis of the file
        self.userFinishedTypingUsernameTimer.start()


    def slotUsernameHistoryTriggered(self, title: str, _message: str) -> None:
        '''Username history button triggered, update the line edit box and update the view of repos'''
        self.ui.lineEditUsername.setText(title)
        self.slotUsernameUpdated()


    def slotUsernameUpdated(self) -> None:
        '''Called when the timer of username update is triggered or when history of username is used'''
        dbg('slotUsernameUpdated')
        username = self.ui.lineEditUsername.text()

        # we use the original repository url for performing the transformation, if possible
        if self.proj_copy is None:
            proj_src = self.proj
        else:
            proj_src = self.proj_copy
        orig_url_dict = { repo.destination: repo.url for repo in proj_src.repos }
        for repo in self.proj.repos:
            orig_url = orig_url_dict[repo.destination]
            if self.ui.radioDoNotAlterUrl.isChecked():
                repo.url = orig_url
            elif self.ui.radioStripUsername.isChecked():
                repo.url = set_username_on_git_url( '', orig_url )
            elif self.ui.radioForceUsername.isChecked():
                repo.url = set_username_on_git_url(username, orig_url)
            else:
                raise AssertionError('No radio for url behavior is checked!')

        self.updateDisplayOfMultigitFile()


    def updateDisplayOfMultigitFile(self) -> None:
        '''Display the project content: items and and in text form'''
        self.ui.textEditProj.setText(str(self.proj))
        # first clear the QTreeWidget
        self.ui.treeWidgetRepoList.clear()
        if len(self.proj.description):
            self.ui.labelProjDesc.setVisible(True)
            self.ui.textEditProjectDesc.setVisible(True)
            self.ui.textEditProjectDesc.setText('<p style="vertical-align: middle">%s</p>' % self.proj.description)
        else:
            self.ui.labelProjDesc.setVisible(False)
            self.ui.textEditProjectDesc.setVisible(False)

        for repo in self.proj.repos:
            # create an item for each repository using MgRepoTreeItem
            # redefined columns for needs of cloning
            item = QTreeWidgetItem(self.ui.treeWidgetRepoList, [repo.destination, repo.head, repo.dest_fullpath, repo.url])
            if repo.url == '':
                item.setText(3, '  -- WARNING: Empty URL --  ')
                f = item.font(3)
                f.setBold(True)
                item.setFont(3, f)
                item.setForeground(3, QColor(Qt.red))
            # adjust columns size
            for i in range(self.ui.treeWidgetRepoList.columnCount()):
                self.ui.treeWidgetRepoList.resizeColumnToContents(i)


    def slotMgitFileHistoryTriggered(self, title: str, _message: str) -> None:
        '''History of Mgit file triggered, open the targeted mgit file'''
        self.ui.lineEditMgitFile.setText(title)
        self.propagateMgitFileUpdated(True, True)


    def slotChooseMgitFile(self) -> None:
        """
        function called when pushing button to select configuration files
        """
        dbg('slotChooseMgitFile')
        # default configuration file
        mgitFile = self.ui.lineEditMgitFile.text()
        # ask user to select configuration file
        mgitFile, _ = QFileDialog.getOpenFileName(self,
                                                        "Select Multigit File",
                                                        mgitFile,
                                                        "Multigit files (*.mgit);;Project Configuration File (*.json)")

        if not mgitFile:
            # dialog canceled, do nothing
            return

        self.ui.lineEditMgitFile.setText(mgitFile)
        self.propagateMgitFileUpdated(True, True)


    def propagateMgitFileUpdated(self, report_file_errors: bool = True, report_parse_errors: bool = True) -> bool:
        '''Parse the json file and udpate the full dialog.

        For report_file_errors, open a warning dialog when the file text is empty or the file does not exist.
        For report_parse_errors, open a warning dialog when the multigit file does not parse correctly.

        Report parse errors
        in a dilog if report_errors is set to True.

        Input data is taken from the widgets:
        * lineEditMgitFile
        * lineEditProjDir

        Return True if everything is valid
        '''
        # clear the project analysis widgets
        self.ui.textEditProj.setText(str(self.proj))
        self.ui.treeWidgetRepoList.clear()

        mgitFile = self.ui.lineEditMgitFile.text()

        # reset proj so that in case of error, all information displayed is empty
        self.proj = ProjectStructure()
        self.proj_copy = None

        if mgitFile == '':
            if report_file_errors:
                QMessageBox.warning(self, 'Multigit file error',
                                    'Multigit file path is empty.')
            self.updateDisplayOfMultigitFile()
            return False

        if not os.path.exists(mgitFile):
            if report_file_errors:
                QMessageBox.warning(None, 'Multigit file error', 'File does not exist: %s' % mgitFile)
            self.updateDisplayOfMultigitFile()
            return False

        if not os.path.isfile(mgitFile):
            if report_file_errors:
                QMessageBox.warning(None, 'Multigit file error', 'This is not a file: %s' % mgitFile)
            self.updateDisplayOfMultigitFile()
            return False

        try:
            self.proj.fill_from_json_file(mgitFile,
                                    pathlib.Path(self.ui.lineEditDestDir.text()))
        except json.JSONDecodeError as exc:
            if report_parse_errors:
                QMessageBox.warning(None, 'Multigit parse error', 'Error during parsing of multigit file %s\n%s' % (mgitFile, str(exc)))
            self.updateDisplayOfMultigitFile()
            return False

        self.proj_copy = ProjectStructure()
        self.proj_copy.fill_from_json_file(mgitFile,
                                      pathlib.Path(self.ui.lineEditDestDir.text()))
        # this will trigger a redisplay of all repos, no need for explicit call to self.updateDisplayOfMultigitFile()
        self.slotUsernameUpdated()

        return True


    def slotDestDirHistoryTriggered(self, title: str, _message: str) -> None:
        '''History of Destination dir triggered, open the targeted destination directory'''
        self.ui.lineEditDestDir.setText(title)
        self.updateDestDir()


    def slotChooseDestDir(self) -> None:
        """
        Called when pushing button to select project destination directory.

        Updates the display of the project file
        :return:
        """
        dbg('slotChooseDestDir')
        # default directory
        destDir = self.ui.lineEditDestDir.text()
        # ask user to select a destination directory for the user
        destDir = QFileDialog.getExistingDirectory(self, 'Choose directory where project will be cloned', destDir)
        if destDir == '':
            return

        self.ui.lineEditDestDir.setText(destDir)
        self.updateDestDir()


    def updateDestDir(self) -> None:
        # set the destination directory in the project object
        destDir = self.ui.lineEditDestDir.text()
        self.proj.set_base_path(pathlib.Path(destDir))
        self.updateDisplayOfMultigitFile()


    def accept(self) -> None:
        '''Called when the dialog is validated'''
        if self.ui.lineEditDestDir.text() == '':
            QMessageBox.warning(None, 'Destination directory error', 'Destination directory is not defined\n')
            return

        targetBaseDir = pathlib.Path(self.ui.lineEditDestDir.text())
        if targetBaseDir.exists() and not targetBaseDir.is_dir():
            QMessageBox.warning(None, 'Destination directory error', 'Destination is not a directory\n')
            return

        if not self.propagateMgitFileUpdated(True, True):
            # there are errors, can not accept the dialog
            return

        repoWithEmptyUrl: List[str] = []
        for repo in self.proj.repos:
            if repo.url == '':
                repoWithEmptyUrl.append(repo.destination)


        if len(repoWithEmptyUrl):
            msg = 'The following repositories have an empty URL. They will not be cloned:\n- ' + '\n- '.join(repoWithEmptyUrl)
            buttonSelectedWarning = QMessageBox.warning(self, 'Repository with empty URL', msg, QMessageBox.Ok | QMessageBox.Abort)
            if buttonSelectedWarning == QMessageBox.Abort:
                # user wants to check what is going on
                return

        repoWithSsh = False
        for repo in self.proj.repos:
            if repo.url.startswith('ssh:'):
                repoWithSsh = True
                break

        if repoWithSsh and self.ui.lineEditUsername.text() == '':
            # ssh cloning requires a username
            msg = 'You are cloning repositories with the ssh:// method, but you forgot to define the username in the dialog.\nSopping the operation'
            buttonSelectedWarning = QMessageBox.warning(self, 'Repository with ssh URL and no user defined', msg, QMessageBox.Abort)
            return


        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_LAST_PROJECT_DIR, self.ui.lineEditDestDir.text())
        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_LAST_MGIT_FILE, self.ui.lineEditMgitFile.text())
        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_CLONE_USERNAME, self.ui.lineEditUsername.text())
        mgc.get_config_instance().save()
        # everyting is fine, accept the dialog
        super().accept()





def runDialogCloneFromMgitFile(window: 'MgMainWindow') -> None:
    """
    This is the dialog to build project structure folders using json configuration file
    :param window: application window
    :return:
    """

    dialogCloneFromMgitFile = MgDialogCloneFromMgitFile(window)

    # display dialog
    dlg_result = dialogCloneFromMgitFile.exec()
    if not dlg_result:
        # Cancel was clicked, do nothing
        return

    # user clicked on OK, display the folder in the main window
    # performing clone

    taskGroups: List[MgExecTaskGroup] = []
    reposToClone = dialogCloneFromMgitFile.proj.repos[:]
    reposToClone.sort(key=lambda r: r.destination)

    skipDir  = []
    toDelDir = []
    alreadyExists = []

    for jsonFileRepo in reposToClone:
        if jsonFileRepo.url == '':
            # the user was informed about this, skip the repo
            continue

        if os.path.exists(jsonFileRepo.dest_fullpath):
            # oh oh, directory exists, let see what user want us to do:
            if dialogCloneFromMgitFile.ui.radioButtonDirExistsSkipDir.isChecked():
                skipDir.append(jsonFileRepo.dest_fullpath)
                # the best way to skip is to do nothing
                continue

            if dialogCloneFromMgitFile.ui.radioButtonDirExistsDelDir.isChecked():
                toDelDir.append(jsonFileRepo.dest_fullpath)

            if dialogCloneFromMgitFile.ui.radioButtonDirExistsGitFail.isChecked():
                # this is the default behavior, so do nothing
                alreadyExists.append(jsonFileRepo.dest_fullpath)


        repoInfo = MgRepoInfo(jsonFileRepo.destination, jsonFileRepo.dest_fullpath, jsonFileRepo.destination)
        taskGroup = MgExecTaskGroup(f'Cloning {repoInfo.name}', repoInfo)
        taskGroup.appendGitTask(f'git clone {repoInfo.name}',
                                ['clone', '--progress', jsonFileRepo.url, jsonFileRepo.dest_fullpath],
                                run_inside_git_repo=False,
                                )

        if len(jsonFileRepo.head):
            # use checkout and not switch because HEAD may point to a branch, tag or commit SHA1
            taskGroup.appendGitTask( f'git checkout {jsonFileRepo.head}', ['checkout', jsonFileRepo.head, '--' ] )

        # if head is empty, the repo is probably empty, so do nothing

        taskGroups.append( taskGroup )

    if len(alreadyExists):
        msg = 'The following repositories already exists:\n'
        msg += '- ' + '\n- '.join(alreadyExists) + '\n\n'
        msg += 'Continue will trigger a git error for these repositories, please consider Deleting them.'
        msgBox = QMessageBox(window)
        msgBox.setWindowTitle('Confirm clone on existing')
        msgBox.setText(msg)
        continueButton = msgBox.addButton('Continue', QMessageBox.AcceptRole)
        delButton = msgBox.addButton('Delete directories', QMessageBox.AcceptRole)
        abortButton = msgBox.addButton(QMessageBox.Abort)
        msgBox.exec()
        buttonSelected = msgBox.clickedButton()

        if buttonSelected == abortButton:
            # user changed its mind ...
            return

        elif buttonSelected == delButton:
            # mark the directories to be deleted
            toDelDir = alreadyExists


    if len(skipDir):
        msg = 'The following repositories already exists and will NOT be cloned (skip behavior):\n'
        msg += '- ' + '\n- '.join(skipDir) + '\n'
        buttonSelectedWarning = QMessageBox.warning(window, 'Confirm skip of repositories', msg, QMessageBox.Ok | QMessageBox.Abort)
        if buttonSelectedWarning == QMessageBox.Abort:
            # user changed its mind ...
            return

    if len(toDelDir):
        # sorting in reverse order allows to remove subdirectories before updirectories
        toDelDir.sort(reverse=True)

        # alreadyExists is set when user chose no special action in the clone dialog, but
        # choose to delete in the second dialog
        if not alreadyExists:
            # the user has not yet validated the list of repos
            msg = 'The following repositories already exists and will be DELETED (delete behavior):\n'
            msg += '- ' + '\n- '.join(toDelDir) + '\n'
            buttonSelectedWarning = QMessageBox.warning(window, 'Confirm deletion of repositories', msg, QMessageBox.Ok | QMessageBox.Abort)
            if buttonSelectedWarning == QMessageBox.Abort:
                # user changed its mind ...
                return

        # kill tgitcache before anything. If it was running, it would prevent deletion of directories
        if sys.platform == 'win32':
            subprocess.call(['taskkill', '/t', '/f', '/im', 'tgitcache.exe'])

        deleteDirList(toDelDir)

        stillExists = [v for v in toDelDir if os.path.exists(v)]
        if len(stillExists):
            # let's try a second time with some delay
            time.sleep(0.5)
            deleteDirList(stillExists)

        results = ''
        stillExists = [v for v in toDelDir if os.path.exists(v)]
        if len(stillExists):
            # let's try a third time with more delay
            time.sleep(0.5)
            results = deleteDirList(stillExists)

        if len(results):
            msg = 'The deletion of repositories returned the following errors:\n' + results
            buttonSelectedWarning = QMessageBox.warning(window, 'Directory deletion error', msg, QMessageBox.Ok | QMessageBox.Abort)
            if buttonSelectedWarning == QMessageBox.Abort:
                # user wants to check what is going on
                return

        if len(stillExists):
            msg = 'Deletion of the following repositories failed:\n'
            msg += '- ' + '\n- '.join(stillExists)
            buttonSelectedWarning = QMessageBox.warning(window, 'Directory not deleted', msg, QMessageBox.Ok | QMessageBox.Abort)
            if buttonSelectedWarning == QMessageBox.Abort:
                # user wants to check what is going on
                return

    # Ensure that clone will work with nested subdirectories
    taskGroups = addPreconditionToEnsureCloneOrderLogic(taskGroups)

    # Give some time between two clones to make sure a clone is fully started
    # before the next one starts, because the next one may create some subdirectories
    # which would then prevent the first clone.
    gitExecWindow = MgExecWindow(window, 0.5)
    gitExecWindow.execTaskGroups('Clone project repositories', taskGroups)

    # show the newly created path in MultiGit
    gitExecWindow.finished.connect( lambda result: window.openDir(dialogCloneFromMgitFile.ui.lineEditDestDir.text()))


class TaskNode:
    def __init__(self, parent: Optional['TaskNode'], name: str, taskGroup: Optional[MgExecTaskGroup]) -> None:
        self.parent = parent
        self.name = name
        self.taskGroup = taskGroup
        self.children: List['TaskNode'] = []


    def addChild(self, parts: List[str], taskGroup: MgExecTaskGroup ) -> 'TaskNode':
        '''Add a child in the graph, in the right node'''
        if len(parts) == 0:
            # we are ready to set the task group
            self.taskGroup = taskGroup
            return self

        for child in self.children:
            if child.name == parts[0]:
                # we have found our children !
                break
        else:
            # we have found no children, create one
            child = TaskNode(self, parts[0], None)
            self.children.append(child)

        # continue the search
        return child.addChild(parts[1:], taskGroup)


def build_taskgroup_dep_graph(taskGroups: List[MgExecTaskGroup]) -> TaskNode:
    '''Build a dependency graph with the nodes and return the top node of the graph'''
    topNode = TaskNode(None, '', None)

    for taskGroup in taskGroups:
        splitter = '/'
        if not splitter in taskGroup.repo.name and '\\' in taskGroup.repo.name:
            splitter = '\\'
        topNode.addChild( taskGroup.repo.name.split(splitter), taskGroup )

    return topNode


def findParentTaskGroup(node: TaskNode) -> Optional[TaskNode]:
    '''Walk the parent of this tasknode until another tasknode is found and return it.

    If no tasknode is found, return None'''
    while node.parent is not None:
        if node.parent.taskGroup is not None:
            return node.parent

        node = node.parent

    return None


def addPreconditionsToTaskGroup(node: TaskNode) -> None:
    '''Using the graph, modify the taskGroup objects in place by adding preconditions
    when one taskGroup depends on another one.'''

    if node.taskGroup is not None:
        # add precondition for this node
        parentTaskGroupNode = findParentTaskGroup(node)

        if parentTaskGroupNode is not None:
            assert parentTaskGroupNode.taskGroup is not None
            node.taskGroup.pre_condition = after_other_taskgroup_is_started_and_dir_exists(parentTaskGroupNode.taskGroup)

    # add preconditions for this nodes children
    for n in node.children:
        addPreconditionsToTaskGroup(n)



def addPreconditionToEnsureCloneOrderLogic(taskGroups: List[MgExecTaskGroup]) -> List[MgExecTaskGroup]:
    '''Generate preconditions so that a subdirectory of a git repo is cloned after the git parent git repo.

    The taskGroup inside the list are modifed in place, with a new precondition.

    Example, with the structure:
    dev/ (git repo)
      + subdev1 (git repo)
        + subdev1_sub1 (git repo)
        + subdev1_sub2 (git repo)
      + subdev2 (git repo)
        + toto
           + subdev2_sub1 (git repo)

    It will generate the following pre-conditions:
    - subdev1 is cloned after dev/ clone has started
    - subdev1_sub1 is cloned after subdev1/ clone has started
    - subdev1_sub2 is cloned after subdev1/ clone has started
    - subdev2 is cloned after dev/ clone has started
    - subdev2_sub1 is cloned after subdev2/ clone has started
    '''
    topNode = build_taskgroup_dep_graph(taskGroups)

    addPreconditionsToTaskGroup(topNode)

    return taskGroups

# TODO
# - rename Structure  to JsonFileStructure
# - rename Repository to JsonFileRepo
# - rename Repository to JsonFilePostCloneCommand





# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_git_push_tag.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from src.mg_repo_tree import MgRepoTree


class Ui_GitPushTag(object):
    def setupUi(self, GitPushTag: QDialog) -> None:
        if not GitPushTag.objectName():
            GitPushTag.setObjectName(u"GitPushTag")
        GitPushTag.resize(965, 492)
        self.verticalLayout_2 = QVBoxLayout(GitPushTag)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.groupBox = QGroupBox(GitPushTag)
        self.groupBox.setObjectName(u"groupBox")
        self.horizontalLayout_2 = QHBoxLayout(self.groupBox)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label = QLabel(self.groupBox)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setPointSize(8)
        self.label.setFont(font)
        self.label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.horizontalLayout_2.addWidget(self.label)

        self.comboBoxTagName = QComboBox(self.groupBox)
        self.comboBoxTagName.setObjectName(u"comboBoxTagName")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(6)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxTagName.sizePolicy().hasHeightForWidth())
        self.comboBoxTagName.setSizePolicy(sizePolicy)
        font1 = QFont()
        font1.setFamily(u"Consolas")
        font1.setPointSize(8)
        self.comboBoxTagName.setFont(font1)
        self.comboBoxTagName.setEditable(True)

        self.horizontalLayout_2.addWidget(self.comboBoxTagName)

        self.widget_2 = QWidget(self.groupBox)
        self.widget_2.setObjectName(u"widget_2")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(3)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.widget_2.sizePolicy().hasHeightForWidth())
        self.widget_2.setSizePolicy(sizePolicy1)

        self.horizontalLayout_2.addWidget(self.widget_2)

        self.widget_4 = QWidget(self.groupBox)
        self.widget_4.setObjectName(u"widget_4")
        sizePolicy2 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy2.setHorizontalStretch(5)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.widget_4.sizePolicy().hasHeightForWidth())
        self.widget_4.setSizePolicy(sizePolicy2)

        self.horizontalLayout_2.addWidget(self.widget_4)


        self.verticalLayout_2.addWidget(self.groupBox)

        self.widget_3 = QWidget(GitPushTag)
        self.widget_3.setObjectName(u"widget_3")

        self.verticalLayout_2.addWidget(self.widget_3)

        self.groupBox_2 = QGroupBox(GitPushTag)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.verticalLayout = QVBoxLayout(self.groupBox_2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.labelRepoSelected = QLabel(self.groupBox_2)
        self.labelRepoSelected.setObjectName(u"labelRepoSelected")
        sizePolicy3 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy3.setHorizontalStretch(1)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.labelRepoSelected.sizePolicy().hasHeightForWidth())
        self.labelRepoSelected.setSizePolicy(sizePolicy3)
        self.labelRepoSelected.setAlignment(Qt.AlignCenter)

        self.horizontalLayout.addWidget(self.labelRepoSelected)

        self.pushButtonAdjustRepoList = QPushButton(self.groupBox_2)
        self.pushButtonAdjustRepoList.setObjectName(u"pushButtonAdjustRepoList")
        sizePolicy4 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy4.setHorizontalStretch(1)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.pushButtonAdjustRepoList.sizePolicy().hasHeightForWidth())
        self.pushButtonAdjustRepoList.setSizePolicy(sizePolicy4)

        self.horizontalLayout.addWidget(self.pushButtonAdjustRepoList)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.treeWidgetRepoList = MgRepoTree(self.groupBox_2)
        self.treeWidgetRepoList.setObjectName(u"treeWidgetRepoList")
        self.treeWidgetRepoList.setAlternatingRowColors(True)
        self.treeWidgetRepoList.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeWidgetRepoList.setRootIsDecorated(False)
        self.treeWidgetRepoList.setUniformRowHeights(True)
        self.treeWidgetRepoList.setItemsExpandable(False)
        self.treeWidgetRepoList.setSortingEnabled(True)

        self.verticalLayout.addWidget(self.treeWidgetRepoList)


        self.verticalLayout_2.addWidget(self.groupBox_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.buttonBox = QDialogButtonBox(GitPushTag)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.horizontalLayout_4.addWidget(self.buttonBox)


        self.verticalLayout_2.addLayout(self.horizontalLayout_4)

#if QT_CONFIG(shortcut)
        self.label.setBuddy(self.comboBoxTagName)
#endif // QT_CONFIG(shortcut)
        QWidget.setTabOrder(self.comboBoxTagName, self.pushButtonAdjustRepoList)
        QWidget.setTabOrder(self.pushButtonAdjustRepoList, self.treeWidgetRepoList)

        self.retranslateUi(GitPushTag)
        self.buttonBox.accepted.connect(GitPushTag.accept)
        self.buttonBox.rejected.connect(GitPushTag.reject)

        QMetaObject.connectSlotsByName(GitPushTag)
    # setupUi

    def retranslateUi(self, GitPushTag: QDialog) -> None:
        GitPushTag.setWindowTitle(QCoreApplication.translate("GitPushTag", u"Git push tag", None))
        self.groupBox.setTitle(QCoreApplication.translate("GitPushTag", u"Git push tag", None))
        self.label.setText(QCoreApplication.translate("GitPushTag", u"Choose tag to push :", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("GitPushTag", u"Targeted repositories", None))
        self.labelRepoSelected.setText(QCoreApplication.translate("GitPushTag", u"3 repositories selected", None))
        self.pushButtonAdjustRepoList.setText(QCoreApplication.translate("GitPushTag", u"Adjust Repository list", None))
        ___qtreewidgetitem = self.treeWidgetRepoList.headerItem()
        ___qtreewidgetitem.setText(4, QCoreApplication.translate("GitPushTag", u"Last Remote Syncho", None));
        ___qtreewidgetitem.setText(3, QCoreApplication.translate("GitPushTag", u"Status", None));
        ___qtreewidgetitem.setText(2, QCoreApplication.translate("GitPushTag", u"Head", None));
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("GitPushTag", u"Git Repo Path", None));
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("GitPushTag", u"hidden", None));
    # retranslateUi




# -*- coding: utf-8 -*-

#    Copyright (c) 2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


################################################################################
## Form generated from reading UI file 'ui_select_repos.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from src.mg_repo_tree import MgRepoTree


class Ui_SelectRepos(object):
    def setupUi(self, SelectRepos: QDialog) -> None:
        if not SelectRepos.objectName():
            SelectRepos.setObjectName(u"SelectRepos")
        SelectRepos.resize(671, 573)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(SelectRepos.sizePolicy().hasHeightForWidth())
        SelectRepos.setSizePolicy(sizePolicy)
        self.verticalLayout = QVBoxLayout(SelectRepos)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.label = QLabel(SelectRepos)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)

        self.horizontalLayout_5.addWidget(self.label)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer)


        self.verticalLayout.addLayout(self.horizontalLayout_5)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalSpacer_7 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_7)

        self.label_3 = QLabel(SelectRepos)
        self.label_3.setObjectName(u"label_3")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy1)

        self.horizontalLayout.addWidget(self.label_3)

        self.lineEditFilterList = QLineEdit(SelectRepos)
        self.lineEditFilterList.setObjectName(u"lineEditFilterList")
        self.lineEditFilterList.setClearButtonEnabled(True)

        self.horizontalLayout.addWidget(self.lineEditFilterList)

        self.horizontalSpacer_8 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_8)

        self.labelLineHidden = QLabel(SelectRepos)
        self.labelLineHidden.setObjectName(u"labelLineHidden")
        sizePolicy1.setHeightForWidth(self.labelLineHidden.sizePolicy().hasHeightForWidth())
        self.labelLineHidden.setSizePolicy(sizePolicy1)

        self.horizontalLayout.addWidget(self.labelLineHidden)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_2)

        self.treeWidgetAvailRepos = MgRepoTree(SelectRepos)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setText(0, u"1");
        self.treeWidgetAvailRepos.setHeaderItem(__qtreewidgetitem)
        self.treeWidgetAvailRepos.setObjectName(u"treeWidgetAvailRepos")
        sizePolicy2 = QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        sizePolicy2.setHorizontalStretch(1)
        sizePolicy2.setVerticalStretch(1)
        sizePolicy2.setHeightForWidth(self.treeWidgetAvailRepos.sizePolicy().hasHeightForWidth())
        self.treeWidgetAvailRepos.setSizePolicy(sizePolicy2)
        self.treeWidgetAvailRepos.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeWidgetAvailRepos.setSortingEnabled(True)
        self.treeWidgetAvailRepos.setAllColumnsShowFocus(True)
        self.treeWidgetAvailRepos.header().setProperty("showSortIndicator", True)

        self.horizontalLayout_2.addWidget(self.treeWidgetAvailRepos)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalSpacer_11 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_11)

        self.horizontalSpacer_9 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_9)

        self.pushButtonMoveAllUp = QPushButton(SelectRepos)
        self.pushButtonMoveAllUp.setObjectName(u"pushButtonMoveAllUp")

        self.horizontalLayout_4.addWidget(self.pushButtonMoveAllUp)

        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_4)

        self.pushButtonMoveUp = QPushButton(SelectRepos)
        self.pushButtonMoveUp.setObjectName(u"pushButtonMoveUp")

        self.horizontalLayout_4.addWidget(self.pushButtonMoveUp)

        self.horizontalSpacer_5 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_5)

        self.pushButtonMoveDown = QPushButton(SelectRepos)
        self.pushButtonMoveDown.setObjectName(u"pushButtonMoveDown")

        self.horizontalLayout_4.addWidget(self.pushButtonMoveDown)

        self.horizontalSpacer_6 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_6)

        self.pushButtonMoveAllDown = QPushButton(SelectRepos)
        self.pushButtonMoveAllDown.setObjectName(u"pushButtonMoveAllDown")

        self.horizontalLayout_4.addWidget(self.pushButtonMoveAllDown)

        self.horizontalSpacer_10 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_10)


        self.verticalLayout.addLayout(self.horizontalLayout_4)

        self.label_2 = QLabel(SelectRepos)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setFont(font)

        self.verticalLayout.addWidget(self.label_2)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_3)

        self.treeWidgetSelectedRepos = MgRepoTree(SelectRepos)
        __qtreewidgetitem1 = QTreeWidgetItem()
        __qtreewidgetitem1.setText(0, u"1");
        self.treeWidgetSelectedRepos.setHeaderItem(__qtreewidgetitem1)
        self.treeWidgetSelectedRepos.setObjectName(u"treeWidgetSelectedRepos")
        sizePolicy3 = QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        sizePolicy3.setHorizontalStretch(1)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.treeWidgetSelectedRepos.sizePolicy().hasHeightForWidth())
        self.treeWidgetSelectedRepos.setSizePolicy(sizePolicy3)
        self.treeWidgetSelectedRepos.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeWidgetSelectedRepos.setSortingEnabled(True)
        self.treeWidgetSelectedRepos.setAllColumnsShowFocus(True)
        self.treeWidgetSelectedRepos.header().setProperty("showSortIndicator", True)

        self.horizontalLayout_3.addWidget(self.treeWidgetSelectedRepos)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.labelRepoSelected = QLabel(SelectRepos)
        self.labelRepoSelected.setObjectName(u"labelRepoSelected")

        self.horizontalLayout_6.addWidget(self.labelRepoSelected)

        self.buttonBox = QDialogButtonBox(SelectRepos)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.horizontalLayout_6.addWidget(self.buttonBox)


        self.verticalLayout.addLayout(self.horizontalLayout_6)


        self.retranslateUi(SelectRepos)
        self.buttonBox.accepted.connect(SelectRepos.accept)
        self.buttonBox.rejected.connect(SelectRepos.reject)

        QMetaObject.connectSlotsByName(SelectRepos)
    # setupUi

    def retranslateUi(self, SelectRepos: QDialog) -> None:
        SelectRepos.setWindowTitle(QCoreApplication.translate("SelectRepos", u"Select repositories", None))
        self.label.setText(QCoreApplication.translate("SelectRepos", u"Available repositories :", None))
        self.label_3.setText(QCoreApplication.translate("SelectRepos", u"Apply filter :", None))
        self.labelLineHidden.setText(QCoreApplication.translate("SelectRepos", u"3 lines hidden", None))
        self.pushButtonMoveAllUp.setText(QCoreApplication.translate("SelectRepos", u"AllDown", None))
        self.pushButtonMoveUp.setText(QCoreApplication.translate("SelectRepos", u"Down", None))
        self.pushButtonMoveDown.setText(QCoreApplication.translate("SelectRepos", u"Up", None))
        self.pushButtonMoveAllDown.setText(QCoreApplication.translate("SelectRepos", u"AllUp", None))
        self.label_2.setText(QCoreApplication.translate("SelectRepos", u"Targeted repositories :", None))
        self.labelRepoSelected.setText(QCoreApplication.translate("SelectRepos", u"TextLabel", None))
    # retranslateUi




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List, Tuple
import logging

from PySide2.QtWidgets import QMessageBox, QAction, QWidget

from src.gui.ui_git_commit import Ui_GitCommit
from src.mg_dialog_utils import MgDialogWithRepoList
from src.mg_exec_window import MgExecWindow
from src.mg_repo_info import MgRepoInfo
import src.mg_config as mgc

logger = logging.getLogger('mg_dialog_git_commit')
dbg = logger.debug


class MgDialogGitCommit(MgDialogWithRepoList):
    ui: Ui_GitCommit

    def __init__(self, parent: QWidget, selectedRepos: List[MgRepoInfo], allRepos: List[MgRepoInfo]) -> None:
        super().__init__(parent,
                         Ui_GitCommit,
                         selectedRepos,
                         allRepos)
        self.ui.buttonHistoryCommit.historyItemTriggered.connect(self.slotHistoryItemTriggered)
        self.fillMenuHistory()


    def fillMenuHistory(self) -> None:
        config = mgc.get_config_instance()
        history = config.lruAsList(mgc.CONFIG_GIT_COMMIT_HISTORY)
        content = []    # type: List[Tuple[str, str]]
        for msg in history:
            if msg.strip() == '':
                # message is empty, do not add action
                continue

            lines = msg.split('\n')
            idx = 0
            # strip empty lines
            while idx < len(lines) and len(lines[idx].strip()) == 0:
                idx += 1

            if idx > len(lines):
                # ok, this message was empty
                continue

            summary = lines[idx].strip()
            content.append((summary, msg))

        self.ui.buttonHistoryCommit.fillHistoryWithTitleAndContent(content)


    def slotHistoryItemTriggered(self, title: str, message: str) -> None:
        '''Triggered when an item from the history menu has been selected'''
        self.ui.textEditCommitMessage.setText(message)


    def accept(self) -> None:
        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_GIT_COMMIT_HISTORY, self.ui.textEditCommitMessage.toPlainText())
        commitMsg = str(self.ui.textEditCommitMessage.toPlainText()).encode('utf8', errors='ignore')
        if len(commitMsg) == 0:
            QMessageBox.warning(self, "Missing commit message",
                                "No commit message specified")
            return

        super().accept()


def runDialogGitCommit(parent: QWidget, selectedRepos: List[MgRepoInfo], allRepos: List[MgRepoInfo]) -> None:
    '''Run commit git command'''
    dbg('runDialogGitCommit')

    dialog = MgDialogGitCommit(parent, selectedRepos, allRepos)
    result = dialog.exec()
    if not result:
        # command execution canceled
        return

    commitMsg = dialog.ui.textEditCommitMessage.toPlainText()
    repoAndListOfGitCmd = []
    desc = "Git Commit"
    for repoInfo in dialog.getTargetedRepoList():
        # commit command
        cmd = ['commit', '-a', '-m', commitMsg]
        repoAndListOfGitCmd.append((desc, repoInfo, [cmd]))

        # push command
        if dialog.ui.checkBoxPushToRemote.isChecked():
            cmd = ['push', '-u', '--progress', '--verbose', 'origin', '{0}'.format(repoInfo.branch)]
            repoAndListOfGitCmd.append(('Push branch %s' % repoInfo.branch,
                                        repoInfo, [cmd]))

    # show window for executing git
    gitExecWindow = MgExecWindow(parent)
    gitExecWindow.execEachRepoWithHisSeqOfGitCommand(desc, repoAndListOfGitCmd)




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List, Any, TYPE_CHECKING, Set
import logging, pathlib, json, os, subprocess, shutil, time, stat

from PySide2.QtWidgets import QDialog, QWidget, QFileDialog, QMessageBox, QTreeWidgetItem, QPushButton, QDialogButtonBox
from PySide2.QtCore import Qt, QTimer, QRegExp
from PySide2.QtGui import QRegExpValidator

if TYPE_CHECKING:
    from src.mg_window import MgMainWindow
from src.gui.ui_apply_mgit_file import Ui_ApplyMgitFile
from src.mg_json_mgit_parser import ProjectStructure
import src.mg_config as mgc
from src.mg_exec_window import MgExecWindow
from src.mg_repo_info import MgRepoInfo
from src.mg_utils import deleteDirList, set_username_on_git_url
from src.mg_dialog_utils import reBranchTagValues
from src.mg_exec_task_item import MgExecTaskGroup
from src.mg_dialog_clone_from_mgit import addPreconditionToEnsureCloneOrderLogic

logger = logging.getLogger('mg_dialog_adjust_project_from_mgit')
dbg = logger.debug
warn = logger.warning


# Cases:
# 1. projConfFile and projDestDir are valid
# 2. update projConfFile
#       2.a -> parse the file -> valid file -> update full GUI
#       2.b -> parse the file -> invalid file -> report error, clear the GUI
# 3. update projDestDir
#       3.a update proj repos destination -> update the display of repo items
#       3.b json file was invalid -> do nothing


class MgDialogApplyMgitFile(QDialog):

    def __init__(self, window: QWidget, destDir: str, currentRepos: List[MgRepoInfo]) -> None:
        super().__init__(window)
        # noinspection PyTypeChecker
        self.setWindowFlags( self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self.ui = Ui_ApplyMgitFile()
        self.ui.setupUi(self)
        self.ui.lineEditDestDir.setText(destDir)
        self.ui.labelProjDesc.setVisible(False)
        self.ui.textEditProjectDesc.setVisible(False)
        self.adjustButton = QPushButton(' Adjust Project ', self)
        self.ui.buttonBox.addButton(self.adjustButton, QDialogButtonBox.AcceptRole)

        self.userFinishedTypingMgitFileTimer = QTimer(self)
        self.userFinishedTypingMgitFileTimer.setSingleShot(True)
        self.userFinishedTypingMgitFileTimer.setInterval(1000)
        self.userFinishedTypingMgitFileTimer.timeout.connect(lambda: self.propagateMgitFileUpdated(False, True))
        self.ui.lineEditMgitFile.textEdited.connect(self.slotMgitFileEdited)

        self.userFinishedTypingUsernameTimer = QTimer(self)
        self.userFinishedTypingUsernameTimer.setSingleShot(True)
        self.userFinishedTypingUsernameTimer.setInterval(1000)
        self.userFinishedTypingUsernameTimer.timeout.connect(self.slotUsernameUpdated)
        self.ui.lineEditUsername.textEdited.connect(self.slotUsernameEdited)
        validator = QRegExpValidator(QRegExp(reBranchTagValues))
        self.ui.lineEditUsername.setValidator(validator)
        self.ui.historyButtonUsername.fillHistory(mgc.get_config_instance().lruAsList(mgc.CONFIG_CLONE_USERNAME))
        self.ui.lineEditUsername.setText(mgc.get_config_instance().lruGetFirst(mgc.CONFIG_CLONE_USERNAME) or '')


        self.currentRepos = currentRepos
        self.proj = ProjectStructure()
        self.reposToClone: List[str] = []

        # set last values
        self.ui.historyButtonMgitFile.fillHistory(mgc.get_config_instance().lruAsList(mgc.CONFIG_LAST_MGIT_FILE))
        self.ui.lineEditMgitFile.setText('')

        # configure QtreeWidget to display list of repositories
        f = self.ui.treeWidgetRepoList.header().font()
        f.setBold(True)
        self.ui.treeWidgetRepoList.header().setFont(f)
        self.ui.treeWidgetRepoList.sortByColumn(0, Qt.AscendingOrder)

        # if both items are filled and exist, updates repos

        # connect buttons to their slot
        self.ui.pushButtonChooseMgitFile.clicked.connect(self.slotChooseMgitFile)
        self.ui.historyButtonMgitFile.historyItemTriggered.connect(self.slotMgitFileHistoryTriggered)
        self.ui.historyButtonUsername.historyItemTriggered.connect(self.slotUsernameHistoryTriggered)

        QTimer.singleShot(500, lambda: self.propagateMgitFileUpdated(False, True))


    def slotMgitFileEdited(self, _text: str) -> None:
        '''Called when user modifies the text'''
        # we wait for 0.5 s then refresh the analysis of the file
        self.userFinishedTypingMgitFileTimer.start()


    def slotUsernameEdited(self, _text: str) -> None:
        '''Called when user modifies the text of the username'''
        # we wait for 0.5 s then refresh the analysis of the file
        self.userFinishedTypingUsernameTimer.start()


    def slotUsernameHistoryTriggered(self, title: str, _message: str) -> None:
        '''Username history button triggered, update the line edit box and update the view of repos'''
        self.ui.lineEditUsername.setText(title)
        self.slotUsernameUpdated()


    def slotUsernameUpdated(self) -> None:
        '''Called when the timer of username update is triggered or when history of username is used'''
        username = self.ui.lineEditUsername.text()
        for repo in self.proj.repos:
            repo.url = set_username_on_git_url(username, repo.url)
        self.updateDisplayOfMultigitFile()


    def updateDisplayOfMultigitFile(self) -> None:
        '''Display the project content: items and and in text form'''
        self.ui.textEditProj.setText(str(self.proj))
        # first clear the QTreeWidget
        self.ui.treeWidgetRepoList.clear()
        if len(self.proj.description):
            self.ui.labelProjDesc.setVisible(True)
            self.ui.textEditProjectDesc.setVisible(True)
            self.ui.textEditProjectDesc.setText('<p style="vertical-align: middle">%s</p>' % self.proj.description)
        else:
            self.ui.labelProjDesc.setVisible(False)
            self.ui.textEditProjectDesc.setVisible(False)

        for repo in self.proj.repos:
            # create an item for each repository using MgRepoTreeItem
            # redefined columns for needs of cloning
            QTreeWidgetItem(self.ui.treeWidgetRepoList, [repo.destination, repo.pretty_head(), repo.dest_fullpath])
            # adjust columns size
            for i in range(self.ui.treeWidgetRepoList.columnCount()):
                self.ui.treeWidgetRepoList.resizeColumnToContents(i)


    def slotMgitFileHistoryTriggered(self, title: str, _message: str) -> None:
        '''History of Mgit file triggered, open the targeted mgit file'''
        self.ui.lineEditMgitFile.setText(title)
        self.propagateMgitFileUpdated(False, True)


    def slotChooseMgitFile(self) -> None:
        """
        function called when pushing button to select configuration files
        """
        dbg('slotChooseMgitFile')
        # default configuration file
        mgitFile = self.ui.lineEditMgitFile.text()
        # ask user to select configuration file
        mgitFile, _ = QFileDialog.getOpenFileName(self,
                                                        "Select Multigit File",
                                                        mgitFile,
                                                        "Multigit files (*.mgit);;Project Configuration File (*.json)")

        if not mgitFile:
            # dialog canceled, do nothing
            return

        self.ui.lineEditMgitFile.setText(mgitFile)
        # this will trigger a redisplay of all repos, no need for explicit call to self.updateDisplayOfMultigitFile()
        self.slotUsernameUpdated()


    def propagateMgitFileUpdated(self, report_file_errors: bool = True, report_parse_errors: bool = True) -> bool:
        '''Parse the json file and udpate the full dialog.

        For report_file_errors, open a warning dialog when the file text is empty or the file does not exist.
        For report_parse_erorrs, open a warning dialog when the multigit file does not parse correctly.

        Report parse errors
        in a dilog if report_errors is set to True.

        Input data is taken from the widgets:
        * lineEditMgitFile
        * lineEditProjDir

        Return True if everything is valid
        '''
        # clear the project analysis widgets
        self.ui.textEditProj.setText(str(self.proj))
        self.ui.treeWidgetRepoList.clear()

        mgitFile = self.ui.lineEditMgitFile.text()

        # reset proj so that in case of error, all information displayed is empty
        self.proj = ProjectStructure()

        if mgitFile == '':
            if report_file_errors:
                QMessageBox.warning(self, 'Multigit file error',
                                    'Multigit file path is empty.')
            self.updateDisplayOfMultigitFile()
            return False

        if not os.path.exists(mgitFile):
            if report_file_errors:
                QMessageBox.warning(None, 'Multigit file error', 'File does not exist: %s' % mgitFile)
            self.updateDisplayOfMultigitFile()
            return False

        if not os.path.isfile(mgitFile):
            if report_file_errors:
                QMessageBox.warning(None, 'Multigit file error', 'This is not a file: %s' % mgitFile)
            self.updateDisplayOfMultigitFile()
            return False

        try:
            self.proj.fill_from_json_file(mgitFile,
                                    pathlib.Path(self.ui.lineEditDestDir.text()))
        except json.JSONDecodeError as exc:
            if report_parse_errors:
                QMessageBox.warning(None, 'Multigit parse error', 'Error during parsing of multigit file %s\n%s' % (mgitFile, str(exc)))
            self.updateDisplayOfMultigitFile()
            return False

        # this will trigger a redisplay of all repos, no need for explicit call to self.updateDisplayOfMultigitFile()
        self.slotUsernameUpdated()

        return True


    def accept(self) -> None:
        '''Called when the dialog is validated'''
        if not self.propagateMgitFileUpdated(True, True):
            # there are errors, can not accept the dialog
            return

        targetRepoSet = set(repo.destination for repo in self.proj.repos)
        currentRepoSet = set(repo.name for repo in self.currentRepos)

        repoNameToCreate = targetRepoSet - currentRepoSet
        repoPathToCreateDirExists = set(repo.dest_fullpath for repo in self.proj.repos
                                            if repo.destination in repoNameToCreate and os.path.exists(repo.dest_fullpath))
        repoNameToRemove = currentRepoSet - targetRepoSet
        repoPathToRemove = [repo.fullpath for repo in self.currentRepos if repo.name in repoNameToRemove]


        if len(repoNameToCreate):
            msg = ''
            msg += 'The following repositories do not exist. They will be cloned:\n - '
            msg += '\n - '.join(repoNameToCreate)
            msg += '\n'
            buttonPressed = QMessageBox.warning(self, 'Cloning repositories', msg, QMessageBox.Ok | QMessageBox.Cancel)
            if buttonPressed != QMessageBox.Ok:
                return

            repoWithEmptyUrl = [
                repo.destination
                for repo in self.proj.repos
                if repo.destination in repoNameToCreate and repo.url == ''
            ]


            if len(repoWithEmptyUrl):
                msg = 'The following repositories to clone have an empty URL. They can not be cloned and will be skipped:\n- ' + '\n- '.join(repoWithEmptyUrl)
                buttonSelectedWarning = QMessageBox.warning(self, 'Repository with empty URL', msg, QMessageBox.Ok | QMessageBox.Abort)
                if buttonSelectedWarning == QMessageBox.Abort:
                    # user wants to check what is going on
                    return

            self.reposToClone = list(sorted(repoNameToCreate))


        if len(repoPathToCreateDirExists):
            msg = ''
            msg += 'The following directories will be removed to allow the git clone:\n - '
            msg += '\n - '.join(repoPathToCreateDirExists)
            msg += '\n'
            buttonPressed = QMessageBox.warning(self, 'Removing directories', msg, QMessageBox.Ok | QMessageBox.Cancel)
            if buttonPressed != QMessageBox.Ok:
                return
            # remove target directory if it already exists to make clone smooth
            result = deleteDirList(repoPathToCreateDirExists)
            if len(result):
                buttonPressed = QMessageBox.warning(self, 'Error during removal of directories',
                                                    'Error during removal of directories: ' + result,
                                                    QMessageBox.Abort | QMessageBox.Ok)
                if buttonPressed != QMessageBox.Ok:
                    return

        if len(repoNameToRemove):
            msg = ''
            msg += 'The following repositories are not in the multigit file and will be removed:\n - '
            msg += '\n - '.join(repoNameToRemove)
            msg += '\n'
            buttonPressed = QMessageBox.warning(self, 'Removing repositories', msg, QMessageBox.Ok | QMessageBox.Cancel)
            if buttonPressed != QMessageBox.Ok:
                return
            # TODO: create a button "delete" for more clarity, and "skip" for even more clarity
            result = deleteDirList(repoPathToRemove)
            if len(result):
                buttonPressed = QMessageBox.warning(self, 'Error during removal of repositories',
                                                    'Error during removal of repositories: ' + result,
                                                    QMessageBox.Abort | QMessageBox.Ok)
                if buttonPressed != QMessageBox.Ok:
                    return

        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_CLONE_USERNAME, self.ui.lineEditUsername.text())
        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_LAST_MGIT_FILE, self.ui.lineEditMgitFile.text())
        mgc.get_config_instance().save()
        # everyting is fine, accept the dialog
        super().accept()


def runDialogApplyMgitFile(window: 'MgMainWindow', baseDir: str, allRepos: List[MgRepoInfo]) -> None:
    """
    This is the dialog to build project structure folders using json configuration file
    :param window: application window
    :return:
    """
    dialogApplyMgitFile = MgDialogApplyMgitFile(window, baseDir, allRepos)

    # display dialog
    dlg_result = dialogApplyMgitFile.exec()
    if not dlg_result:
        # Cancel was clicked, do nothing
        return

    # user clicked on OK, display the folder in the main window
    # performing project adjustment

    taskGroups: List[MgExecTaskGroup] = []
    reposToAdjust =  dialogApplyMgitFile.proj.repos[:]
    reposToAdjust.sort(key=lambda r: r.destination)

    for repoDictInfo in reposToAdjust:
        repoInfo = MgRepoInfo(repoDictInfo.destination, repoDictInfo.dest_fullpath, repoDictInfo.destination)
        if repoDictInfo.destination in dialogApplyMgitFile.reposToClone:
            if repoDictInfo.url == '':
                # user has already been warned about this
                continue

            # we must clone these, no checkout them
            taskGroup = MgExecTaskGroup(f'Cloning and checkouting repository {repoInfo.name}', repoInfo)
            taskGroup.appendGitTask(f'git clone {repoInfo.name}',
                                    ['clone', '--progress', repoDictInfo.url, repoDictInfo.dest_fullpath])
            taskGroup.appendGitTask(f'git checkout {repoDictInfo.head}',
                                    ['checkout', repoDictInfo.head])
        else:
            seq_of_git_cmd = [
                ['fetch', '--prune', '--tags'],
                ['checkout', repoDictInfo.head],
            ]
            taskGroup = MgExecTaskGroup(f'Fetching and checkouting repository {repoInfo.name}', repoInfo)
            taskGroup.appendGitTask(f'git fetch {repoInfo.name}',
                                    ['fetch', '--prune', '--tags'])
            taskGroup.appendGitTask(f'git checkout {repoDictInfo.head}',
                                    ['checkout', repoDictInfo.head, '--']
                                    )

        taskGroups.append( taskGroup )


    # Ensure that clone will work with nested subdirectories
    taskGroups = addPreconditionToEnsureCloneOrderLogic(taskGroups)

    gitExecWindow = MgExecWindow(window)
    gitExecWindow.execTaskGroups('Adjusting project repositories', taskGroups)
    gitExecWindow.finished.connect(lambda _result: window.actionRefreshAll.trigger())


# TODO:
# use colors just like in main repo display branch/tag/commit
# use colors just like in main repo display branch/tag/commit



# -*- coding: utf-8 -*-

#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


################################################################################
## Form generated from reading UI file 'ui_repo_properties.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_RepoProperties(object):
    def setupUi(self, RepoProperties: QDialog) -> None:
        if not RepoProperties.objectName():
            RepoProperties.setObjectName(u"RepoProperties")
        RepoProperties.resize(844, 804)
        self.gridLayout = QGridLayout(RepoProperties)
        self.gridLayout.setObjectName(u"gridLayout")
        self.label = QLabel(RepoProperties)
        self.label.setObjectName(u"label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.lineEditDir = QLineEdit(RepoProperties)
        self.lineEditDir.setObjectName(u"lineEditDir")
        self.lineEditDir.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEditDir, 0, 1, 1, 1)

        self.label_3 = QLabel(RepoProperties)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)

        self.lineEditFullPath = QLineEdit(RepoProperties)
        self.lineEditFullPath.setObjectName(u"lineEditFullPath")
        self.lineEditFullPath.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEditFullPath, 1, 1, 1, 1)

        self.buttonCopyPath = QPushButton(RepoProperties)
        self.buttonCopyPath.setObjectName(u"buttonCopyPath")

        self.gridLayout.addWidget(self.buttonCopyPath, 1, 2, 1, 1)

        self.label_4 = QLabel(RepoProperties)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.gridLayout.addWidget(self.label_4, 2, 0, 1, 1)

        self.lineEditUrl = QLineEdit(RepoProperties)
        self.lineEditUrl.setObjectName(u"lineEditUrl")
        self.lineEditUrl.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEditUrl, 2, 1, 1, 1)

        self.buttonCopyUrl = QPushButton(RepoProperties)
        self.buttonCopyUrl.setObjectName(u"buttonCopyUrl")

        self.gridLayout.addWidget(self.buttonCopyUrl, 2, 2, 1, 1)

        self.label_5 = QLabel(RepoProperties)
        self.label_5.setObjectName(u"label_5")
        self.label_5.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.gridLayout.addWidget(self.label_5, 3, 0, 1, 1)

        self.lineEditHead = QLineEdit(RepoProperties)
        self.lineEditHead.setObjectName(u"lineEditHead")
        self.lineEditHead.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEditHead, 3, 1, 1, 1)

        self.label_7 = QLabel(RepoProperties)
        self.label_7.setObjectName(u"label_7")
        self.label_7.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.gridLayout.addWidget(self.label_7, 4, 0, 1, 1)

        self.lineEditStatus = QLineEdit(RepoProperties)
        self.lineEditStatus.setObjectName(u"lineEditStatus")
        self.lineEditStatus.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEditStatus, 4, 1, 1, 1)

        self.label_6 = QLabel(RepoProperties)
        self.label_6.setObjectName(u"label_6")
        self.label_6.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.gridLayout.addWidget(self.label_6, 5, 0, 1, 1)

        self.lineEditRemoteBranch = QLineEdit(RepoProperties)
        self.lineEditRemoteBranch.setObjectName(u"lineEditRemoteBranch")
        self.lineEditRemoteBranch.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEditRemoteBranch, 5, 1, 1, 1)

        self.label_9 = QLabel(RepoProperties)
        self.label_9.setObjectName(u"label_9")
        self.label_9.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.gridLayout.addWidget(self.label_9, 6, 0, 1, 1)

        self.lineEditRemoteSynchro = QLineEdit(RepoProperties)
        self.lineEditRemoteSynchro.setObjectName(u"lineEditRemoteSynchro")
        self.lineEditRemoteSynchro.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEditRemoteSynchro, 6, 1, 1, 1)

        self.label_10 = QLabel(RepoProperties)
        self.label_10.setObjectName(u"label_10")

        self.gridLayout.addWidget(self.label_10, 7, 0, 1, 1)

        self.lineEditTags = QLineEdit(RepoProperties)
        self.lineEditTags.setObjectName(u"lineEditTags")

        self.gridLayout.addWidget(self.lineEditTags, 7, 1, 1, 1)

        self.label_8 = QLabel(RepoProperties)
        self.label_8.setObjectName(u"label_8")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(3)
        sizePolicy1.setHeightForWidth(self.label_8.sizePolicy().hasHeightForWidth())
        self.label_8.setSizePolicy(sizePolicy1)
        self.label_8.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.gridLayout.addWidget(self.label_8, 8, 0, 1, 1)

        self.textEditLastCommit = QTextEdit(RepoProperties)
        self.textEditLastCommit.setObjectName(u"textEditLastCommit")
        sizePolicy2 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(3)
        sizePolicy2.setHeightForWidth(self.textEditLastCommit.sizePolicy().hasHeightForWidth())
        self.textEditLastCommit.setSizePolicy(sizePolicy2)
        self.textEditLastCommit.setReadOnly(True)
        self.textEditLastCommit.setAcceptRichText(False)

        self.gridLayout.addWidget(self.textEditLastCommit, 8, 1, 1, 1)

        self.label_13 = QLabel(RepoProperties)
        self.label_13.setObjectName(u"label_13")
        sizePolicy1.setHeightForWidth(self.label_13.sizePolicy().hasHeightForWidth())
        self.label_13.setSizePolicy(sizePolicy1)

        self.gridLayout.addWidget(self.label_13, 9, 0, 1, 1)

        self.textEditDiffSummary = QTextEdit(RepoProperties)
        self.textEditDiffSummary.setObjectName(u"textEditDiffSummary")
        sizePolicy2.setHeightForWidth(self.textEditDiffSummary.sizePolicy().hasHeightForWidth())
        self.textEditDiffSummary.setSizePolicy(sizePolicy2)
        self.textEditDiffSummary.setReadOnly(True)

        self.gridLayout.addWidget(self.textEditDiffSummary, 9, 1, 1, 1)

        self.buttonBox = QDialogButtonBox(RepoProperties)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Ok)
        self.buttonBox.setCenterButtons(False)

        self.gridLayout.addWidget(self.buttonBox, 10, 1, 1, 1)


        self.retranslateUi(RepoProperties)
        self.buttonBox.accepted.connect(RepoProperties.accept)
        self.buttonBox.rejected.connect(RepoProperties.reject)

        QMetaObject.connectSlotsByName(RepoProperties)
    # setupUi

    def retranslateUi(self, RepoProperties: QDialog) -> None:
        RepoProperties.setWindowTitle(QCoreApplication.translate("RepoProperties", u"Git Repository properties", None))
        self.label.setText(QCoreApplication.translate("RepoProperties", u"Directory :", None))
        self.label_3.setText(QCoreApplication.translate("RepoProperties", u"Full repository path :", None))
#if QT_CONFIG(tooltip)
        self.buttonCopyPath.setToolTip(QCoreApplication.translate("RepoProperties", u"Copy full path to clipboard", None))
#endif // QT_CONFIG(tooltip)
        self.buttonCopyPath.setText("")
        self.label_4.setText(QCoreApplication.translate("RepoProperties", u" URL :", None))
#if QT_CONFIG(tooltip)
        self.buttonCopyUrl.setToolTip(QCoreApplication.translate("RepoProperties", u"Copy URL to clipboard", None))
#endif // QT_CONFIG(tooltip)
        self.buttonCopyUrl.setText("")
        self.label_5.setText(QCoreApplication.translate("RepoProperties", u"Head :", None))
        self.label_7.setText(QCoreApplication.translate("RepoProperties", u"Status :", None))
        self.label_6.setText(QCoreApplication.translate("RepoProperties", u"Remote Branch :", None))
        self.label_9.setText(QCoreApplication.translate("RepoProperties", u"Remote Synchro :", None))
        self.label_10.setText(QCoreApplication.translate("RepoProperties", u"Tags for this commit :", None))
        self.label_8.setText(QCoreApplication.translate("RepoProperties", u"Last commit :", None))
        self.label_13.setText(QCoreApplication.translate("RepoProperties", u"Diff summary :", None))
    # retranslateUi

